#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Input stream parsing framework
============================================================================

Dragonfly's generic input stream parser is built around the concept of
parser elements, each of which can consume a certain form of input.  These
parser elements can be constructed into a hierarchy describing the expected
type of input they are meant to process.

"""


#---------------------------------------------------------------------------

import string
import re
import locale
import logging

from six import string_types, text_type, binary_type


class ParserError(Exception):
    pass


#---------------------------------------------------------------------------

class Parser(object):

    def __init__(self, parser_element, log=None):
        self._parser_element = parser_element
        self._log = log

    def parse(self, input, must_finish=True):
        state = State(input, log=self._log)
        generator = self._parser_element.parse(state)
        for _ in generator:
            if not must_finish or state.finished():
                # Parse complete, return result.
                node = state.build_parse_tree()
                return node.value()
        # Failed to parse.
        return None

    def parse_node(self, input, must_finish=True):
        state = State(input, log=self._log)
        generator = self._parser_element.parse(state)
        for result in generator:
            if not must_finish or state.finished():
                # Parse complete, return result.
                node = state.build_parse_tree()
                return node
        # Failed to parse.
        return None

    def parse_multiple(self, input, must_finish=True):
        state = State(input, log=self._log)
        generator = self._parser_element.parse(state)

        values = []
        for result in generator:
            if not must_finish or state.finished():
                values.append(state.build_parse_tree().value())
        return values

#       return [state.build_parse_tree().value()
#           for result in generator
#           if not must_finish or state.finished()]


class State(object):

    #-----------------------------------------------------------------------
    # Methods for initialization.

    def __init__(self, data, index=0, log=None):
        self._data = data
        self._index = index
        self._log = log
        self._log_debug = self._log.isEnabledFor(logging.DEBUG) if self._log else False
        self._stack = []
        self._depth = 0
        self._previous_index = None

    def initialize_decoding(self):
        self._depth = 0
        self._stack = []
        self._previous_index = None

    def __repr__(self):
        return self.position_string()

    #-----------------------------------------------------------------------
    # Methods for accessing recognition content.

    def position_string(self, width=20, width_before=6):
        mark, continuation = ">>", "."
        width -= 2

        # Locate edges.
        i1 = self._index - width_before
        i2 = self._index - width_before + width - len(mark)

        # Account for edge effects.
        if i1 < 0:
            i2 = min(i2 - i1, len(self._data))
            i1 = 0
        elif i2 > len(self._data):
            width_before += i2 - len(self._data)
            i1 = max(i1 - i2 + len(self._data), 0)
            i2 = len(self._data)

        # Collect character representations.
        before = [repr(c)[1:-1] for c in self._data[i1:self._index]]
        after = [repr(c)[1:-1] for c in self._data[self._index:i2]]

        # Build before string.
        if i1 == 0 and len("".join(before)) <= width_before:
            before = '"' + "".join(before)
        else:
            while before and len("".join(before)) > width_before - 2:
                before.pop(0)
            before = "".join(before)
            before = '%s"%s' % (continuation * (width_before - len(before)), before)

        # Build after string.
        width_after = width - len(before) - len(mark)
        if i2 == len(self._data) and len("".join(after)) <= width_after:
            after = "".join(after) + '"'
        else:
            while after and len("".join(after)) > width_after - 2:
                after.pop()
            after = "".join(after)
            after = '%s"%s' % (after, continuation * (width_after - len(after)))

        # Build and return result.
        return '%s%s%s' % (before, mark, after)

    #-----------------------------------------------------------------------
    # Methods for parsing of input.

    def remaining(self):
        return len(self._data) - self._index

    def finished(self):
        return (len(self._data) - self._index) <= 0

    def peek(self, length):
        if self._index == len(self._data):
            return None
        elif self._index + length > len(self._data):
            length = len(self._data) - self._index
        return self._data[self._index : self._index + length]

    def next(self, length):
        if self._index == len(self._data):
            return None
        elif self._index + length > len(self._data):
            length = len(self._data) - self._index
        self._index += length
        return self._data[self._index - length : self._index]

    def build_parse_tree(self):
        root, index = self._build_parse_node(0, None)
        return root

    def _build_parse_node(self, index, parent):
        frame = self._stack[index]
        node = Node(parent, frame.actor, self._data,
                    frame.begin, frame.end, frame.depth,
                    frame.value)
        if parent:
            parent.add_child(node)
        index += 1
        while index < len(self._stack):
            if self._stack[index].depth != frame.depth + 1:
                break
            child, index = self._build_parse_node(index, node)
        return node, index

    #-----------------------------------------------------------------------
    # Methods for tracking parsing of input.

    class _Frame(object):
        __slots__ = ("depth", "actor", "begin", "end", "value")

        def __init__(self, depth, actor, begin):
            self.depth, self.actor, self.begin = depth, actor, begin
            self.value, self.end = None, None

    def decode_attempt(self, element):
        # assert isinstance(element, ParserElementBase)
        self._depth += 1
        self._stack.append(State._Frame(self._depth, element, self._index))
        if self._log_debug:
            self._log_step(element, "attempt")

    def decode_retry(self, element):
        # assert isinstance(element, ParserElementBase)
        frame = self._get_frame_from_actor(element)
        self._depth = frame.depth
        if self._log_debug:
            self._log_step(element, "retry")

    def decode_rollback(self, element):
        # assert isinstance(element, ParserElementBase)
        frame = self._get_frame_from_depth()
        if not frame or frame.actor != element:
            raise ParserError("Parser decoding stack broken")
        if frame is self._stack[-1]:
            # Last parser on the stack, rollback.
            self._index = frame.begin
        else:
            raise ParserError("Parser decoding stack broken")
        if self._log_debug:
            self._log_step(element, "rollback")

    def decode_success(self, element, value=None):
        # assert isinstance(element, ParserElementBase)
        if self._log_debug:
            self._log_step(element, "success")
        frame = self._get_frame_from_depth()
        if not frame or frame.actor != element:
            raise ParserError("Parser decoding stack broken.")
        frame.end = self._index
        frame.value = value
        self._depth -= 1

    def decode_failure(self, element):
        # assert isinstance(element, ParserElementBase)
        frame = self._stack.pop()
        self._index = frame.begin
        self._depth = frame.depth
        if self._log_debug:
            self._log_step(element, "failure")
        self._depth -= 1

    def _get_frame_from_depth(self):
        for frame in reversed(self._stack):
            if frame.depth == self._depth:
                return frame
        return None

    def _get_frame_from_actor(self, actor):
        for frame in reversed(self._stack):
            if frame.actor is actor:
                return frame
        return None

    def _log_step(self, parser, message):
        indent = "   " * self._depth
        output = "%s%s: %s" % (indent, message, parser)
        self._log.debug(output)
        if self._index != self._previous_index:
            self._previous_index = self._index
            output = "%s -- Decoding State: '%s'" % (indent, str(self))
            self._log.debug(output)


#---------------------------------------------------------------------------

class Node(object):

    __slots__ = ("parent", "children", "actor",
                 "data", "begin", "end", "depth", "success_value")

    def __init__(self, parent, actor, data, begin, end, depth, value):
        self.parent, self.actor, self.data = parent, actor, data
        self.begin, self.end, self.depth = begin, end, depth
        self.success_value = value
        self.children = []

    def __repr__(self):
        if isinstance(self.data, binary_type):
            data = self.data.decode(locale.getpreferredencoding())
        else:
            data = text_type(self.data)
        return "Node: %s, %s" % (self.actor, data)

    def add_child(self, child):
        self.children.append(child)

    def match(self):
        return self.data[self.begin : self.end]

    def value(self):
        if self.success_value is not None:
            return self.success_value
        else:
            return self.actor.value(self)

    def get_children(self, name=None, actor_type=None, shallow=True):
        found = []
        for child in self.children:
            if (    (not name or child.actor.name == name)
                and (not actor_type or isinstance(child.actor, actor_type))):
                found.append(child)
            if shallow and child.actor.name:
                continue
            found.extend(child.get_children(name=name, actor_type=actor_type,
                                            shallow=shallow))
        return found

    def get_child(self, name=None, actor_type=None, shallow=True):
        for child in self.children:
            if (    (not name or child.actor.name == name)
                and (not actor_type or isinstance(child.actor, actor_type))):
                return child
            if shallow and child.actor.name:
                continue
            result = child.get_child(name=name, actor_type=actor_type,
                                     shallow=shallow)
            if result:
                return result
        return None

    def pretty_string(self, indent=""):
        if not self.children:
            return "%s%s" % (indent, str(self))
        else:
            return ("%s%s\n" % (indent, str(self)) +
                    "\n".join([n  .pretty_string(indent + "  ")
                               for n in self.children]))


#---------------------------------------------------------------------------
# Element base class.

class ParserElementBase(object):

    def __init__(self, name=None):
        self._name = name

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def _str(self, argument):
        if self._name:
            return "%s(%s)" % (self._name, argument)
        else:
            return "%s(%s)" % (self.__class__.__name__, argument)

    def __repr__(self):
        return self._str("...")

    name = property(lambda self: self._name,
                    doc="Read-only access to name attribute.")

    def _get_children(self):
        return ()
    children = property(lambda self: self._get_children(),
                        doc="Generalized access to child elements.")

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def parse(self, state):
        raise NotImplementedError("Call to virtual method parse()"
                                  " in class %s" % self)

    def value(self, node):
        return node.match()


#---------------------------------------------------------------------------
# Basic structural element classes.

class Sequence(ParserElementBase):

    def __init__(self, children=(), name=None):
        ParserElementBase.__init__(self, name)
        self._children = children

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        return self._str("%d children" % len(self._children))

    def _get_children(self):
        return self._children

    @property
    def child_list(self):
        return self._children

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def parse(self, state):
        state.decode_attempt(self)

        # Special case for an empty sequence.
        if not self._children:
            state.decode_success(self)
            yield state
            state.decode_retry(self)
            state.decode_failure(self)
            return

        # Attempt to walk a path through the entire sequence of children
        #  so that each one decodes successfully.
        path = [self._children[0].parse(state)]
        while path:
            # Allow the last child to attempt decoding.
            try:
                next(path[-1])
            except StopIteration:
                # Last child failed to decode, remove from path to
                #  allowed the one-before-last child to reattempt.
                path.pop()
            else:
                # Last child successfully decoded.
                if len(path) < len(self._children):
                    # Sequence not yet complete, append the next child.
                    path.append(self._children[len(path)].parse(state))
                else:
                    # Sequence complete, all children decoded successfully.
                    state.decode_success(self)
                    yield state
                    state.decode_retry(self)

        # Sequence of children could not all decode successfully: failure.
        state.decode_failure(self)
        return

    def value(self, node):
        return [c.value() for c in node.children]


#---------------------------------------------------------------------------

class Repetition(ParserElementBase):

    def __init__(self, child, min=1, max=None, name=None):
        ParserElementBase.__init__(self, name)

        if not isinstance(child, ParserElementBase):
            raise TypeError("Child %s must be a"
                            " ParserElementBase instance." % child)
        self._child = child
        self._min = min
        self._max = max
        self._greedy = True

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        return self._str("")

    def _get_children(self):
        return (self._child, )

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def parse(self, state):
        state.decode_attempt(self)

        # Attempt to walk a path through the entire sequence of children
        #  so that each one decodes successfully.
        path = [self._child.parse(state)]
        while path:
            # Allow the last child to attempt decoding.
            try:
                next(path[-1])
            except StopIteration:
                # Last child failed to decode, remove from path to
                #  allow the one-before-last child to reattempt.
                path.pop()

                if self._greedy and len(path) >= self._min:
                    # Path long enough, yield success.
                    state.decode_success(self)
                    yield state
                    state.decode_retry(self)

            else:
                # Last child successfully decoded.
                if not self._max or len(path) < self._max:
                    if not self._greedy and len(path) >= self._min:
                        # Path long enough, yield success.
                        state.decode_success(self)
                        yield state
                        state.decode_retry(self)
                    # Path not too long, append the next child.
                    path.append(self._child.parse(state))
                else:
                    # Path length at maximum, yield success.
                    state.decode_success(self)
                    yield state
                    state.decode_retry(self)

        # Sequence of children could not all decode successfully: failure.
        state.decode_failure(self)
        return

    def value(self, node):
        return [c.value() for c in node.children]


#---------------------------------------------------------------------------

class Alternative(ParserElementBase):

    def __init__(self, children=(), name=None):
        ParserElementBase.__init__(self, name)
        self._children = list(children)

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        return self._str("%d children" % len(self._children))

    def _get_children(self):
        return tuple(self._children)

    @property
    def child_list(self):
        return self._children

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def parse(self, state):
        state.decode_attempt(self)

        # Special case for an empty list of alternatives.
        if not self._children:
            state.decode_success(self)
            yield state
            state.decode_retry(self)
            state.decode_failure(self)
            return

        # Iterate through the children.
        for child in self._children:

            # Iterate through this child's possible decoding states.
            for result in child.parse(state):
                state.decode_success(self)
                yield state
                state.decode_retry(self)

            # Rollback to the alternative's original state, so that the
            #  next child starts decoding without interference from the
            #  previous child.
            state.decode_rollback(self)

        # None of the children could decode successfully: failure.
        state.decode_failure(self)
        return

    def value(self, node):
        return node.children[0].value()


#---------------------------------------------------------------------------

class Optional(ParserElementBase):

    def __init__(self, child, greedy=True, name=None):
        ParserElementBase.__init__(self, name)

        if not isinstance(child, ParserElementBase):
            raise TypeError("Child %s must be a"
                            " ParserElementBase instance." % child)
        self._child = child
        self._greedy = greedy

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        return self._str("")

    def _get_children(self):
        return (self._child, )

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def parse(self, state):
        state.decode_attempt(self)

        # If in greedy mode, allow the child to decode before.
        if self._greedy:
            for result in self._child.parse(state):
                state.decode_success(self)
                yield state
                state.decode_retry(self)

            # Rollback decoding so that the null-decode can be yielded.
            state.decode_rollback(self)

        # Yield the null-decode possibility.
        state.decode_success(self)
        yield state
        state.decode_retry(self)

        # If not in greedy mode, allow the child to decode after.
        if not self._greedy:
            # Rollback decoding after null-decode.  Perhaps not absolutely
            #  necessary, but should be done for good form.
            state.decode_rollback(self)

            for result in self._child.parse(state):
                state.decode_success(self)
                yield state
                state.decode_retry(self)

        # No more decoding possibilities available, failure.
        state.decode_failure(self)
        return

    def value(self, node):
        if node.children:
            return node.children[0].value()
        else:
            return None


#===========================================================================

class String(ParserElementBase):
    r"""
        Parser element for static strings.

        Usage examples:
        >>> def parse_multiple(element, input):
        ...     parser = Parser(element)
        ...     return parser.parse_multiple(input, must_finish=False)
        >>> parse_multiple(String("foo"), "foo")
        ['foo']
        >>> parse_multiple(String("foo"), "foobar")
        ['foo']
        >>> parse_multiple(String("\n\t "), "\n\t foo")
        ['\n\t ']
        >>> parse_multiple(String("foo"), "bar")
        []
        >>> parse_multiple(String("foo"), " foo")
        []

    """

    def __init__(self, string, name=None):
        ParserElementBase.__init__(self, name)
        self._string = string

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        return self._str("%s" % self._string)

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def parse(self, state):
        state.decode_attempt(self)

        # Check if the next characters match the target string.
        if state.next(len(self._string)) == self._string:
            state.decode_success(self)
            yield state
            state.decode_retry(self)

        state.decode_failure(self)
        return


#---------------------------------------------------------------------------

class CharacterSeries(ParserElementBase):
    """
    Class for parsing from a character series or pattern.
    """

    def __init__(self, set, optional=False, exclude=False, name=None,
                 pattern=None):
        ParserElementBase.__init__(self, name)
        self._set = set
        self._optional = optional
        self._exclude = exclude
        self._pattern = pattern

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        return self._str("%s" % self._set)

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def char_matches(self, c):
        if self._set:
            return c in self._set
        elif self._pattern:
            return bool(self._pattern.match(c))
        else:
            return False

    def parse(self, state):
        state.decode_attempt(self)

        # Gobble as many valid characters as possible.
        count = 0
        if self._exclude:
            while (not state.finished() and not
                    self.char_matches(state.peek(1))):
                state.next(1)
                count += 1
        else:
            while (not state.finished() and
                    self.char_matches(state.peek(1))):
                state.next(1)
                count += 1

        if self._optional or count > 0:
            state.decode_success(self)
            yield state
            state.decode_retry(self)

        state.decode_failure(self)
        return


#---------------------------------------------------------------------------

class Choice(Alternative):

    def __init__(self, choices, name=None):
        choice_pairs = []
        choice_elements = []
        for key, value in choices.items():
            if isinstance(key, string_types):
                element = String(key)
            elif isinstance(key, ParserElementBase):
                element = key
            else:
                raise TypeError("Invalid choice key: %r" % key)
            choice_pairs.append((element, value))
            choice_elements.append(element)
        self._choice_pairs = choice_pairs
        Alternative.__init__(self, choice_elements, name)

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        return self._str("%d choices" % len(self._choice_pairs))

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def value(self, node):
        actor = node.children[0].actor
        for element, value in self._choice_pairs:
            if actor == element:
                return value
        raise ParserError("Invalid child element: %s" % actor)


#---------------------------------------------------------------------------

class Whitespace(CharacterSeries):

    def __init__(self, optional=False, name=None):
        set = string.whitespace
        CharacterSeries.__init__(self, set, optional=optional, name=name)

    def __repr__(self):
        return self._str("")


class Letters(CharacterSeries):
    """
    Class for parsing ascii and non-ascii letter characters.
    """
    def __init__(self, name=None):
        pattern = re.compile(r"\w", re.UNICODE)
        CharacterSeries.__init__(self, None, name=name, pattern=pattern)

    def __repr__(self):
        return self._str("")

    def char_matches(self, c):
        if c.isdigit():
            return False
        else:
            return super(Letters, self).char_matches(c)


class Alphanumerics(CharacterSeries):
    """
    Class for parsing ascii and non-ascii letter and digit characters.
    """
    def __init__(self, name=None):
        pattern = re.compile(r"\w", re.UNICODE)
        CharacterSeries.__init__(self, None, name=name, pattern=pattern)

    def __repr__(self):
        return self._str("")


#---------------------------------------------------------------------------

class QuotedStringContent(ParserElementBase):

    escape_char    = "\\"
    valid_chars    = string.printable
    escaped_chars  = {
                      "n":  "\n",
                      "t":  "\t",
                     }
    valid_char_pattern = re.compile(r"\w", re.UNICODE)

    def __init__(self, delimiter_string, name=None):
        self.delimiter_string = delimiter_string
        ParserElementBase.__init__(self, name)

    def valid_char(self, c):
        """
        Check whether a character is a valid. Valid characters include all
        Unicode alphanumeric characters and any characters in
        self.valid_chars.
        :type c: str
        :rtype: bool
        """
        if c in self.valid_chars:
            return True
        else:
            return bool(self.valid_char_pattern.match(c))

    def parse(self, state):
        state.decode_attempt(self)

        # Gobble as many valid characters as possible.
        characters = []
        while not state.finished():
            next_char = state.peek(1)

            # If next_char escapes, look forward past it and
            #  handle the escaped character.
            if next_char == self.escape_char:
                # If there's nothing after the escaped char, then
                #  we can't gobble the escape character itself.
                #  So we're done.
                if state.remaining() < 2:
                    break
                escaped_char = state.peek(2)[1]
                if escaped_char in self.escaped_chars:
                    # The escaped character has a special transformation.
                    next_char = self.escaped_chars[escaped_char]
                else:
                    # The escaped character is not special,
                    #  so simply unescape it.
                    next_char = escaped_char
                    if not self.valid_char(next_char):
                        break
                state.next(1)  # Gobble self.escape_char.

            # If self.delimiter_string is next, then we're done.
            elif state.peek(len(self.delimiter_string)) == self.delimiter_string:
                break

            # If next_char is not acceptable, don't gobble it.
            elif not self.valid_char(next_char):
                break

            state.next(1)  # Gobble next_char.
            characters.append(next_char)

        value = "".join(characters)
        state.decode_success(self, value)
        yield state
        state.decode_retry(self)
        state.decode_failure(self)
        return


#---------------------------------------------------------------------------

class QuotedString(Alternative):
    """
        Parser element for quoted strings.

        Simple usage showing default delimiters:
        >>> parser = Parser(QuotedString())
        >>> parser.parse("''")
        ''
        >>> parser.parse("'Hello world!'")
        'Hello world!'
        >>> parser.parse('"Hello world!"')
        'Hello world!'

        Special characters within the quoted string can be escaped.
        >>> parser.parse(r'"Hello \"world\"!"')  #doctest: +SKIP
        Note back to get these escaping examples to run using doctest
        they must include double escapes, which are obviously not required
        in actual code.
        >>> parser.parse(r'"Hello \\"world\\"!"')
        'Hello "world"!'
        >>> print(parser.parse(r'"Hello \\\\ \\"world\\"!\\nGoodbye \\'universe\\'..."'))
        Hello \ "world"!
        Goodbye 'universe'...

        This element supports asymmetric open-close delimiters:
        >>> parser = Parser(QuotedString([("[[", "]]")]))
        >>> parser.parse("[[Hello world!]]")
        'Hello world!'
        >>> parser.parse("[[Hello world!]] Goodbye.", must_finish=False)
        'Hello world!'

    """

    # Open-close delimiter pairs.
    default_delimiters = (
                          ('"', '"'),
                          ("'", "'"),
                         )

    def __init__(self, delimiters=default_delimiters, name=None):
        self.delimiters = delimiters
        children = []
        for open_delimiter, close_delimiter in self.delimiters:
            delimiter_children = (
                                  String(open_delimiter),
                                  QuotedStringContent(close_delimiter),
                                  String(close_delimiter),
                                 )
            child = Sequence(delimiter_children)
            children.append(child)
        Alternative.__init__(self, children, name=name)

    def value(self, node):
        return node.children[0].children[1].value()


#---------------------------------------------------------------------------

class UnsignedInteger(CharacterSeries):
    """
        Parser element for unsigned integer literals.

        Simple usage examples:
        >>> from dragonfly.parser import Parser
        >>> parser = Parser(UnsignedInteger())
        >>> parser.parse("0")
        0
        >>> parser.parse("1234")
        1234
        >>> parser.parse("0001234")
        1234

    """

    digits_set = string.digits

    def __init__(self, name=None):
        CharacterSeries.__init__(self, self.digits_set, name)

    def value(self, node):
        return int(CharacterSeries.value(self, node))


#---------------------------------------------------------------------------

class Integer(Sequence):
    """
        Parser element for quoted strings.

        Simple usage examples:
        >>> from dragonfly.parser import Parser
        >>> parser = Parser(Integer())
        >>> parser.parse("0")
        0
        >>> parser.parse("+0")
        0
        >>> parser.parse("-000")
        0
        >>> parser.parse("1234")
        1234
        >>> parser.parse("+001234")
        1234
        >>> parser.parse("-001234")
        -1234

    """

    sign_strings = {
                    "+":  1,
                    "-": -1,
                   }
    digits_set = UnsignedInteger.digits_set

    def __init__(self, name=None):
        children = (
                    Optional(Choice(self.sign_strings)),
                    UnsignedInteger(),
                   )
        Sequence.__init__(self, children, name)

    def value(self, node):
        sign = node.children[0].value() or 1
        magnitude = node.children[1].value()
        return sign * magnitude


#---------------------------------------------------------------------------

class Float(Sequence):
    """
        Parser element for decimal fraction literals.

        Usage examples:
        >>> from dragonfly.parser import Parser
        >>> parser = Parser(Float())
        >>> parser.parse("0.0")
        0.0
        >>> parser.parse(".000")
        0.0
        >>> parser.parse("-.0")
        0.0
        >>> parser.parse("1.0")
        1.0
        >>> parser.parse("-1.0")
        -1.0
        >>> parser.parse("-1.75")
        -1.75

    """

    separator_string = "."

    def __init__(self, name=None):
        digits = CharacterSeries(set=string.digits)
        children = (
                    Optional(Alternative([
                                         Integer(),
                                         Choice(Integer.sign_strings, name="sign_only"),
                                        ])),
                    String(self.separator_string),
                    UnsignedInteger(),
                   )
        Sequence.__init__(self, children=children, name=name)

    def value(self, node):
        sign_only_node = node.get_child(name="sign_only")
        if sign_only_node:
            if float(sign_only_node.data) >= 0:
                integer_part = ""
            else:
                integer_part = "-"
        else:
            integer_part = node.children[0].value() or 0
        fractional_part = node.children[2].value()
        return float("%s.%d" % (integer_part, fractional_part))


#---------------------------------------------------------------------------

def print_matches(node, indent=""):
    if not indent:
        print("Nodes:")
    print(indent, "%s: %r %s" % (node.actor.__class__.__name__,
                                 node.match(), id(node.parent)))
    [print_matches(c, indent + "   ") for c in node.children]


def print_values(node, indent=""):
    if not indent:
        print("Values:")
    print(indent, "%s: %r %s" % (node.actor.__class__.__name__, node.value(),
                                 id(node.parent)))
    [print_values(c, indent + "   ") for c in node.children]



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

from pynput.mouse import Controller, Button

from ._base import BaseButtonEvent, MoveEvent


# Initialise a pynput mouse controller.
_controller = Controller()


#---------------------------------------------------------------------------
# Functions and event delegate for getting and setting the cursor position.

def get_cursor_position():
    return _controller.position


def set_cursor_position(x, y):
    _controller.position = (x, y)
    return True


class MoveEventDelegate(object):

    @classmethod
    def get_position(cls):
        return get_cursor_position()

    @classmethod
    def set_position(cls, x, y):
        return set_cursor_position(x, y)


# Set MoveEvent's delegate. This allows us to set platform-specific
# functions for getting and setting the cursor position without having
# to do a lot of sub-classing for no good reason.
MoveEvent.delegate = MoveEventDelegate()

#---------------------------------------------------------------------------
# Pynput mouse button and wheel up/down flags.


def get_button(name):
    # If possible, get a Button from the name.
    # This allows button and wheel flags to work, at least partially, on
    # multiple platforms.
    return getattr(Button, name, None)


PLATFORM_BUTTON_FLAGS = {
    # ((button, event_type), down)
    # The inner pair is used here and below to be compatible with the
    # original Windows flags.
    "left":   (((get_button("left"), 0), 1),  # down
               ((get_button("left"), 0), 0)),  # up
    "right":  (((get_button("right"), 0), 1),
               ((get_button("right"), 0), 0)),
    "middle": (((get_button("middle"), 0), 1),
               ((get_button("middle"), 0), 0)),

    # We call these "four" and "five" because Windows calls them that.
    # These buttons typically behave as browser back and forward media keys.
    "four": (((get_button("button8"), 0), 1),
             ((get_button("button8"), 0), 0)),
    "five": (((get_button("button9"), 0), 1),
             ((get_button("button9"), 0), 0)),
}

PLATFORM_WHEEL_FLAGS = {
    # ((button, event_type), scroll_count)
    "wheelup": ((get_button("scroll_up"), 1), 3),
    "stepup": ((get_button("scroll_up"), 1), 1),
    "wheeldown": ((get_button("scroll_down"), 1), 3),
    "stepdown": ((get_button("scroll_down"), 1), 1),
    "wheelright": ((get_button("scroll_right"), 1), 3),
    "stepright": ((get_button("scroll_right"), 1), 1),
    "wheelleft": ((get_button("scroll_left"), 1), 3),
    "stepleft": ((get_button("scroll_left"), 1), 1),
}


#---------------------------------------------------------------------------
# pynput event classes.


class ButtonEvent(BaseButtonEvent):

    def execute(self, window):
        for ((button, event_type), flag) in self._flags:
            # Check if the button is unknown.
            if button is None:
                event_type_s = "button" if event_type == 0 else "scroll"
                raise ValueError("Unsupported %s event" % event_type_s)

            if event_type == 0:  # Button press event
                if flag:
                    _controller.press(button)
                else:
                    _controller.release(button)
            elif event_type == 1:  # Scroll event
                _controller.click(button, flag)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Tools for testing element classes
============================================================================

"""

import logging
from six import string_types

from dragonfly              import *
from ..test                 import TestError, RecognitionFailure
from ..test.infrastructure  import Unique


#===========================================================================

class ElementTester(Grammar):

    _log = logging.getLogger("test.element")
    _NotSet = Unique("NoRecognition")

    class _ElementTestRule(Rule):
        exported = True
        def process_recognition(self, node):
            self.grammar._process_recognition(node)

    #-----------------------------------------------------------------------

    def __init__(self, element, engine=None):
        Grammar.__init__(self, self.__class__.__name__, engine=engine)
        rule = self._ElementTestRule("rule", element)
        self.add_rule(rule)

    def recognize(self, words):
        if isinstance(words, string_types):
            words = words.split()

        if not self.loaded:
            self._log.debug("Loading ElementTester instance.")
            self.load()
            unload_after_recognition = True
        else:
            unload_after_recognition = False

        self._recognized_value = self._NotSet
        try:
            # Make this grammar exclusive; this *probably* avoids other
            #  grammars from being active and receiving the mimicked
            #  recognition.
            self.set_exclusiveness(True)

            # Mimic recognition.
            try:
                self.engine.mimic(words)
            except MimicFailure as e:
                self._recognized_value = RecognitionFailure
            except Exception as e:
                self._log.exception("Exception within recognition: %s" % (e,))
                raise

        except Exception as e:
            self._log.exception("Exception during recognition: %s" % (e,))
            raise
        finally:
            if unload_after_recognition:
                try:
                    self.unload()
                except Exception as e:
                    raise TestError("Failed to unload grammar: %s" % e)

        # If recognition was successful but this grammar did not
        #  receive the recognition callback, then apparently some other
        #  grammar hijacked it; raise a TestError to signal this
        #  undesired situation.
        if self._recognized_value == self._NotSet:
            self._log.error(u"Recognition hijacked. (Words: %s)" % (words,))
            raise TestError(u"Recognition hijacked. (Words: %s)" % (words,))

        # Return the value of the element after recognition.
        return self._recognized_value

    def _process_recognition(self, node):
        element_node = node.children[0]
        self._recognized_value = element_node.value()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import time

from dragonfly.actions.action_base import ActionError
from dragonfly.windows import monitors


#---------------------------------------------------------------------------
# Functions and event delegate for getting and setting the cursor position.

def get_cursor_position():
    message = ("Getting the cursor position is not implemented on this "
               "platform!")
    raise NotImplementedError(message)


def set_cursor_position(x, y):
    message = ("Setting the cursor position is not implemented on this "
               "platform!")
    raise NotImplementedError(message)


class MoveEventDelegate(object):

    @classmethod
    def get_position(cls):
        return get_cursor_position()

    @classmethod
    def set_position(cls, x, y):
        return set_cursor_position(x, y)


#---------------------------------------------------------------------------
# Mouse button and wheel up/down flags.
# These flags should be overridden for each supported platform.

PLATFORM_BUTTON_FLAGS = {
    "left":   ((1, 1),   # down
               (1, 0)),  # up
    "right":  ((2, 1),
               (2, 0)),
    "middle": ((3, 1),
               (3, 0)),
    "four":   ((4, 1),
               (4, 0)),
    "five":   ((5, 1),
               (5, 0)),
}


PLATFORM_WHEEL_FLAGS = {
    "wheelup": (1, 120),
    "stepup": (1, 40),
    "wheeldown": (1, -120),
    "stepdown": (1, -40),
    "wheelright": (2, 120),
    "stepright": (2, 40),
    "wheelleft": (2, -120),
    "stepleft": (2, -40),
}


#---------------------------------------------------------------------------
# Event classes.


class EventBase(object):

    def execute(self, window):
        pass


class MoveEvent(EventBase):

    # Set the event delegate. This allows us to set platform-specific
    # functions for getting and setting the cursor position without having
    # to do a lot of sub-classing for no good reason.
    delegate = MoveEventDelegate()

    def __init__(self, from_left, horizontal, from_top, vertical):
        self.from_left = from_left
        self.horizontal = horizontal
        self.from_top = from_top
        self.vertical = vertical
        EventBase.__init__(self)

    def _move_relative(self, rectangle):
        if self.from_left:  horizontal = rectangle.x1
        else:               horizontal = rectangle.x2
        if isinstance(self.horizontal, float):
            distance = self.horizontal * rectangle.dx
        else:
            distance = self.horizontal
        horizontal += distance

        if self.from_top:   vertical = rectangle.y1
        else:               vertical = rectangle.y2
        if isinstance(self.vertical, float):
            distance = self.vertical * rectangle.dy
        else:
            distance = self.vertical
        vertical += distance

        self._move_mouse(horizontal, vertical)

    def _get_position(self):
        return self.delegate.get_position()

    def _move_mouse(self, horizontal, vertical):
        self.delegate.set_position(horizontal, vertical)


class MoveWindowEvent(MoveEvent):

    def execute(self, window):
        self._move_relative(window.get_position())


class MoveScreenEvent(MoveEvent):

    def execute(self, window):
        # Move the mouse relative to the first monitor's rectangle.
        # Do nothing if there are no monitors in the list.
        if monitors:
            self._move_relative(monitors[0].rectangle)


class BaseButtonEvent(EventBase):

    def __init__(self, *flags):
        EventBase.__init__(self)
        self._flags = flags

    def execute(self, window):
        message = ("Mouse button events are not implemented for this "
                   "platform!")
        raise NotImplementedError(message)


class MoveRelativeEvent(MoveEvent):

    def __init__(self, horizontal, vertical):
        MoveEvent.__init__(self, None, None, None, None)
        self.horizontal = horizontal
        self.vertical = vertical

    def execute(self, window):
        position = self._get_position()
        if not position:
            raise ActionError("Failed to retrieve cursor position.")
        horizontal = position[0] + self.horizontal
        vertical   = position[1] + self.vertical
        self._move_mouse(horizontal, vertical)


class PauseEvent(EventBase):

    def __init__(self, interval):
        EventBase.__init__(self)
        self._interval = interval

    def execute(self, window):
        time.sleep(self._interval)



"""This module contains OS-independent functionality *which does not depend on
other Dragonfly functionality*. Functionality which does depend on Dragonfly
classes such as actions should go into controller.py. This separation is done to
make it easier in the future to extract this functionality as a general Python
library independent of Dragonfly.
"""


import regex
import enum
import logging


_log = logging.getLogger("accessibility")


class CursorPosition(enum.Enum):
    """The cursor position relative to a range of text."""

    BEFORE = 1
    """The position before the text."""

    AFTER = 2
    """The position after the text."""


class TextQuery(object):
    """A query to match a range of text."""

    def __init__(self,
                 start_phrase="",
                 start_relative_position=None,
                 start_relative_phrase="",
                 through=False,
                 end_phrase="",
                 end_relative_position=None,
                 end_relative_phrase=""):
        if not (end_phrase or end_relative_phrase):
            raise ValueError("Must specify end (relative) phrase")
        if not through and (start_phrase or start_relative_phrase):
            raise ValueError("Must specify 'through' if start is specified.")
        if start_relative_phrase and not start_relative_position:
            raise ValueError("Must specify before or after for start_relative_phrase")
        if end_relative_phrase and not end_relative_position:
            raise ValueError("Must specify before or after for end_relative_phrase")

        self.start_phrase = start_phrase
        """The phrase at the start of the match."""

        self.start_relative_position = start_relative_position
        """Whether to match before or after the :attr:`start_relative_phrase`."""

        self.start_relative_phrase = start_relative_phrase
        """The phrase to match before or after at the start."""

        self.through = through
        """True if matching from a start point to the end phrase."""

        self.end_phrase = end_phrase
        """The phrase at the end of the match (or the sole phrase)."""

        self.end_relative_position = end_relative_position
        """Whether to match before or after the :attr:`end_relative_phrase`."""

        self.end_relative_phrase = end_relative_phrase
        """The phrase to match before or after at the end."""

    def __repr__(self):
        return str(dict([(k, v) for (k, v) in self.__dict__.items() if v]))


class TextInfo(object):
    """Information about a range of text."""

    def __init__(self, start, end, text, start_coordinates=None, end_coordinates=None):
        self.start = start
        self.end = end
        self.text = text
        self.start_coordinates = start_coordinates
        self.end_coordinates = end_coordinates


def _phrase_to_regex(phrase):
    # Treat whitespace between words as meaning anything other than alphanumeric
    # characters.
    pattern = r"[^\w--_]+".join(regex.escape(word) for word in phrase.split())
    # Treat spaces at the beginning or end of the phrase as matching any
    # whitespace character. This makes it easy to select stuff like non-breaking
    # space, which occurs frequently in browsers.
    # TODO Support newlines. Note that these are frequently implemented as
    # separate text nodes in the accessibility tree, so the obvious
    # implementation would not work well.
    if phrase == " ":
        pattern = r"\s"
    else:
        if phrase.startswith(" "):
            pattern = r"\s" + pattern
        if phrase.endswith(" "):
            pattern = pattern + r"\s"
    # Only match at boundaries of alphanumeric sequences if the phrase ends
    # are alphanumeric.
    if regex.search(r"^[\w--_]", phrase, regex.VERSION1 | regex.UNICODE):
        pattern = r"(?<![\w--_])" + pattern
    if regex.search(r"[\w--_]$", phrase, regex.VERSION1 | regex.UNICODE):
        pattern = pattern + r"(?![\w--_])"
    return pattern


def _find_text(query, expanded_text, cursor_offset):
    # Convert query to regex with a single matching group. Note that this used
    # to be simpler but buggy, and is now is complicated because lookbehind and
    # lookahead regular expressions must be fixed-width, so we have to limit
    # their usage and use a matching group instead, which requires matching up
    # the parens surrounding the group.
    pattern = r""

    # Add the start phrases, if present.
    if query.start_phrase or query.start_relative_phrase:
        if query.start_relative_phrase and query.start_relative_position == CursorPosition.AFTER:
            pattern += _phrase_to_regex(query.start_relative_phrase)
            if query.start_phrase:
                pattern += r"[^\w--_]*"
        pattern += r"(" + _phrase_to_regex(query.start_phrase)
        if query.start_relative_phrase and query.start_relative_position == CursorPosition.BEFORE:
            if query.start_phrase:
                pattern += r"[^\w--_]*"
            pattern += _phrase_to_regex(query.start_relative_phrase)
        pattern += r".*?"

    # Add the end phrases.
    if query.end_relative_phrase and query.end_relative_position == CursorPosition.AFTER:
        pattern += _phrase_to_regex(query.end_relative_phrase)
        if query.end_phrase:
            pattern += r"[^\w--_]*"
    # Add the initial "(" if we haven't already.
    if not (query.start_phrase or query.start_relative_phrase):
        pattern += r"("
    pattern += _phrase_to_regex(query.end_phrase) + r")"
    if query.end_relative_phrase and query.end_relative_position == CursorPosition.BEFORE:
        if query.end_phrase:
            pattern += r"[^\w--_]*"
        pattern += _phrase_to_regex(query.end_relative_phrase)

    # Find all matches.
    matches = regex.finditer(pattern,
                             expanded_text,
                             regex.IGNORECASE | regex.VERSION1 | regex.UNICODE,
                             overlapped=True)
    ranges = [(match.start(1), match.end(1)) for match in matches]
    if not ranges:
        _log.warning("Not found: %s" % query)
        return None
    start_with_cursor = (query.through and not (query.start_phrase or query.start_relative_phrase))
    if cursor_offset is None:
        _log.warning("Cursor not found.")
        if start_with_cursor:
            _log.warning("Cannot perform cursor-relative query without cursor.")
            return None
        else:
            # Pick arbitrary match (the first one).
            return ranges[0]
    else:
        # Find nearest match.
        range = min(ranges, key=lambda x: abs((x[0] + x[1]) / 2 - cursor_offset))
        if start_with_cursor:
            if range[0] < cursor_offset:
                return (range[0], cursor_offset)
            else:
                return (cursor_offset, range[1])
        else:
            return range


def _get_focused_text(context):
    if not context.focused:
        _log.warning("Nothing is focused.")
        return None
    focused_text = context.focused.as_text()
    if not focused_text:
        _log.warning("Focused element is not text.")
        return None
    return focused_text


def get_cursor_offset(controller):
    def closure(context):
        focused_text = _get_focused_text(context)
        if not focused_text:
            return None
        return focused_text.cursor
    return controller.run_sync(closure)


def set_cursor_offset(controller, offset):
    def closure(context):
        focused_text = _get_focused_text(context)
        if not focused_text:
            return None
        focused_text.set_cursor(offset)
    controller.run_sync(closure)


def get_text_info(controller, query):
    _log.debug("Getting text info: %s" % query)
    def closure(context):
        focused_text = _get_focused_text(context)
        if not focused_text:
            return None
        nearest = _find_text(query, focused_text.expanded_text, focused_text.cursor)
        if not nearest:
            return None
        text_info = TextInfo(nearest[0], nearest[1],
                             focused_text.expanded_text[nearest[0]:nearest[1]])
        start_box = focused_text.get_bounding_box(nearest[0])
        end_box = focused_text.get_bounding_box(nearest[1] - 1)
        # Ignore offscreen coordinates. Google Docs returns these, and we handle
        # it here to avoid further trouble.
        if start_box.x < 0 or start_box.y < 0 or end_box.x < 0 or end_box.y < 0:
            _log.warning("Text selection points were offscreen, ignoring.")
        else:
            text_info.start_coordinates = start_box.x, start_box.y + start_box.height / 2
            text_info.end_coordinates = end_box.x + end_box.width, end_box.y + end_box.height / 2
        return text_info
    return controller.run_sync(closure)


def move_cursor(controller, query, position):
    """Moves the cursor before or after the provided phrase."""

    _log.info("Moving cursor %s phrase: %s" % (position.name, query))
    def closure(context):
        focused_text = _get_focused_text(context)
        if not focused_text:
            return False
        nearest = _find_text(query, focused_text.expanded_text, focused_text.cursor)
        if not nearest:
            return False
        focused_text.set_cursor(nearest[0] if position is CursorPosition.BEFORE else nearest[1])
        _log.info("Moved cursor")
        return True
    return controller.run_sync(closure)


def select_text(controller, query):
    _log.info("Selecting text: %s" % query)
    def closure(context):
        focused_text = _get_focused_text(context)
        if not focused_text:
            return False
        nearest = _find_text(query, focused_text.expanded_text, focused_text.cursor)
        if not nearest:
            return False
        focused_text.select_range(*nearest)
        _log.info("Selected text")
        return True
    return controller.run_sync(closure)


def is_editable_focused(controller):
    """Returns true if an editable object is focused."""

    def closure(context):
        return context.focused and context.focused.is_editable()
    return controller.run_sync(closure)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Configuration toolkit
============================================================================

Dragonfly's configuration toolkit makes it very easy to store
program data in a separate file from the main program logic.  It uses
a three-phase *setup -- load -- use* system:

 - *setup* -- a Config object is created and its structure and
   default contents are defined.
 - *load* -- a separate file containing the user's configuration
   settings is looked for and, if found, its values are loaded
   into the Config object.
 - *use* -- the program directly accesses the configuration
   through easy Config object attributes.

This configuration toolkit uses the following three classes:

 - :class:`Config` -- a collection of configuration settings, grouped
   within one or more sections
 - :class:`Section` -- a group of items and/or subsections
 - :class:`Item` -- a single configuration setting

Usage example
----------------------------------------------------------------------------

The main program using Dragonfly's configuration toolkit would
normally look something like this: ::

    from dragonfly import Config, Section, Item

    # *Setup* phase.
    # This defines a configuration object with the name "Example
    #  configuration".  It contains one section with the title
    #  "Test section", which has two configuration items.  Both
    #  these items have a default value and a docstring.
    config                 = Config("Example configuration")
    config.test            = Section("Test section")
    config.test.fruit      = Item("apple", doc="Must eat fruit.")
    config.test.color      = Item("blue", doc="The color of life.")

    # *Load* phase.
    # This searches for a file with the same name as the main program,
    #  but with the extension ".py" replaced by ".txt".  It is also
    #  possible to explicitly specify the configuration file's path.
    #  See Config.load() for more details.
    config.load()

    # *Use* phase.
    # The configuration values can now be accessed through the
    #  configuration object as follows.
    print "The color of life is", config.test.color
    print "You must eat an %s every day" % config.test.fruit

The configuration defined above is basically complete.  Every
configuration item has a default value and can be accessed by
the program.  But if the user would like to modify some or all
of these settings, he can do so in an external configuration file
without modifying the main program code.

This external configuration file is interpreted as Python code.
This gives its author powerful tools for determining the desired
configuration settings.  However, it will usually consist merely
of variable assignments. The configuration file for the program
above might look something like this: ::

    # Test section
    test.fruit = "banana"   # Bananas have more potassium.
    test.color = "white"    # I like light colors.


Example command modules
----------------------------------------------------------------------------

The configuration toolkit is utilized in a number of command modules in the
*t4ngo/dragonfly-modules* repository, available on GitHub.  See
:ref:`Related Resources: Command modules <RefCommandModulesList>`.


Implementation details
----------------------------------------------------------------------------

This configuration toolkit makes use of Python's special methods
for setting and retrieving object attributes.  This makes it much
easier to use, as there is no need to use functions such as
*value = get_config_value("item_name")*; instead the configuration
values are immediately accessible as Python objects.  It also allows
for more extensive error checking; it is for example trivial to
implement custom *Item* classes which only allow specific values
or value types, such as integers, boolean values, etc.


Configuration class reference
----------------------------------------------------------------------------

"""

import sys
import os.path
import inspect
import textwrap
import traceback

import logging


#---------------------------------------------------------------------------
# Config processing modes:
#  - initializing -- Config is being set up and populated.
#  - loading -- Config setup is complete, loading from user file.
#  - done -- Config initialization and loading complete, ready to use.

_init, _load, _done = range(3)


#---------------------------------------------------------------------------
# Config class; this manages a command module's configuration.

class Config(object):
    """
        Configuration class for storing program settings.

        Constructor argument:
         - *name* (*str*) --
           the name of this configuration object.

        This class can contain zero or more :class:`Section` instances,
        each of which can contain zero or more :class:`Item` instances.
        It is these items which store the actual configuration settings.
        The sections merely divide the items up into groups, so that
        different configuration topics can be split for easy readability.

  """

    _configs_by_name = {}
    _log = logging.getLogger("config")


    #-----------------------------------------------------------------------

    @classmethod
    def get_by_name(cls, name):
        try:
            return cls._configs_by_name[name]
        except KeyError:
            return None

    @classmethod
    def get_instances(cls):
        instances = list(cls._configs_by_name.items())
        instances.sort()
        return [instance for name, instance in instances]


    #-----------------------------------------------------------------------

    def __init__(self, name):
        set_ = object.__setattr__
        set_(self, "name", name)
        set_(self, "_sections", {})
        set_(self, "_sections_list", [])
        set_(self, "_mode", _init)
        set_(self, "module_path", None)
        set_(self, "config_path", None)

        Config._configs_by_name[name] = self

    def _set_mode(self, mode):
        object.__setattr__(self, "_mode", mode)
        for n, s in self._sections_list:
            s._set_mode(mode)

    def __getattr__(self, name):
        if name in self._sections:
            return self._sections[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if self._mode == _init:
            if isinstance(value, Section):
                self._sections[name] = value
                self._sections_list.append((name, value))
            else: raise TypeError("Invalid type %s, expecting Section."
                                  % type(value))
        else: raise AttributeError(name)


    def load(self, path=None):
        """
            Load the configuration file at the given *path*, or
            look for a configuration file associated with the calling
            module.

             - *path* (*str*, default: *None*) --
               path to the configuration file to load.  If *None*,
               then a path is generated from the calling module's
               file name by replacing its extension with ".txt".

            If the *path* is a file, it is loaded.  On the other hand,
            if it does not exist or is not a file, nothing is loaded
            and this configuration's defaults remain in place.

        """
        self._set_mode(_load)
        if not path:
            caller_frame = inspect.currentframe().f_back
            caller_file = caller_frame.f_globals["__file__"]
            module_base, module_ext = os.path.splitext(caller_file)
            path = module_base + ".txt"
            if module_ext in (".pyc", ".pyo"):
                module_ext = ".py"
            object.__setattr__(self, "module_path", module_base + module_ext)
        object.__setattr__(self, "config_path", path)

        if os.path.exists(path):
            namespace = self.load_from_file(path)
        else:
            namespace = None
        self._set_mode(_done)

        return namespace

    def load_from_file(self, path):
        namespace = dict(self._sections)
        for name, section in self._sections_list:
            section.update_namespace(namespace)

        try:
            with open(path) as file:
                exec(compile(file.read(), path, 'exec'), namespace)
#        except ConfigError, e:
        except Exception as e:
            print("exception:", e)
            t, v, tb = sys.exc_info()
            frames = traceback.extract_tb(tb)
            relevant_frames = []
            error_line = "<unknown>"
            include_all = False
            for frame in frames:
                filename, line, function, text = frame
                print("frame:", frame)

                if not include_all:
                    file1 = os.path.basename(filename)
                    file2 = os.path.basename(path)
                    if file1 == file2:
                        include_all = True
                        error_line = line

                if include_all:
                    relevant_frames.append(frame)

            self._log.error("An error occurred in the %s file at line %s."
                            % (path, error_line))
            self._log.error("The error message was: %s" % e)
            formatted = traceback.format_list(relevant_frames)
            lines = "\n".join(formatted).splitlines()
            for line in lines:
                self._log.error("    " + line)

        return namespace

    _comment_indent = " "*20
    _comment_wrapper = textwrap.TextWrapper(
                                   width=70,
                                   break_long_words=False,
                                   initial_indent=_comment_indent+"# ",
                                   subsequent_indent=_comment_indent+"#  ",
                                  )

    def _format_file_head(self):
        header = "Dragonfly config for %(config_name)s" % {"config_name": self.name}
        lines = self._comment_wrapper.wrap(header)
        lines.insert(0, "#")
        lines.append("#")
        lines.append("")
        return lines

    def _format_section_head(self, name, section, names):
        width = 76
        header = "#--- %s " % section.doc
        length = width - len(header)
        if length > 0: header += "-" * length
        return [header, ""]

    def _format_item(self, name, item, names):
        path = ".".join(list(names) + [name])
        data = {
                "doc":      item.doc,
                "default":  item.default,
                "value":    item.value,
                "path":     path,
               }
        lines = []
        lines.append("%(path)s = %(value)r" % data)
        if item.doc:
            header = "%(doc)s" % data
            lines.extend(self._comment_wrapper.wrap(header))
            lines.extend(self._comment_wrapper.wrap("Default: %(default)r" % data))
        lines.append("")
        return lines

    def generate_config_file(self, path=None):
        """
            Create a configuration file containing this
            configuration object's current settings.

             - *path* (*str*, default: *None*) --
               path to the configuration file to load.  If *None*,
               then a path is generated from the calling module's
               file name by replacing its extension with ".txt".

        """
        # Config file header.
        output = self._format_file_head()

        # Iterates through the sections.
        stack = [iter(self._sections_list)]
        names = []
        while stack:
            # Try to retrieve the next section from the top of the stack.
            try:
                section_name, section = next(stack[-1])
                names.append(section_name)
            except StopIteration:
                # No more subsections, remove section from top of stack.
                # Check for names needed for when the root Config is popped.
                stack.pop()
                if names: names.pop()
                continue

            # Generate section head.
            section_head = self._format_section_head(section_name, section, names)
            output.extend(section_head)

            # Output section items.
            for item_name, item in section._items_list:
                item_output = self._format_item(item_name, item, names)
                output.extend(item_output)

            # Push section onto the stack.
            stack.append(iter(section._sections_list))

        if not path:
            caller_frame = inspect.currentframe().f_back
            caller_file = caller_frame.f_globals["__file__"]
            module_base = os.path.splitext(caller_file)[0]
            path = module_base + ".txt"

        f = open(path, "w")
        f.write("\n".join(output))
        f.close()
        f = None


#---------------------------------------------------------------------------
# Section class; this represents a section within a Config hierarchy.

class Section(object):
    """
        Section of a configuration for grouping items.

        Constructor argument:
         - *doc* (*str*) --
           the name of this configuration section.

        A section can contain zero or more subsections and zero or more
        configuration items.

    """

    def __init__(self, doc):
        set_ = object.__setattr__
        set_(self, "doc", doc)
        set_(self, "_items", {})
        set_(self, "_items_list", [])
        set_(self, "_sections", {})
        set_(self, "_sections_list", [])
        set_(self, "_mode", _init)

    def _set_mode(self, mode):
        object.__setattr__(self, "_mode", mode)
        for n, s in self._sections_list:
            s._set_mode(mode)

    def __getattr__(self, name):
        if name in self._items:        return self._items[name].value
        elif name in self._sections:   return self._sections[name]
        else:                          raise AttributeError(name)

    def __setattr__(self, name, value):
        if self._mode == _init:
            if isinstance(value, Item):
                self._items[name] = value
                self._items_list.append((name, value))
            elif isinstance(value, Section):
                self._sections[name] = value
                self._sections_list.append((name, value))
            else:
                raise TypeError("Invalid type %s, expecting Item or Section."
                                % type(value))
        elif self._mode == _load:
            if name in self._items:    self._items[name].value = value
            else:                      raise AttributeError(name)
        else:
            raise AttributeError(name)

    def update_namespace(self, namespace):
        for name, item in self._items_list:
            item.update_namespace(namespace)
        for name, section in self._sections_list:
            section.update_namespace(namespace)


#---------------------------------------------------------------------------
# Item classes which represent config values.

class Item(object):
    """
        Configuration item for storing configuration settings.

        Constructor arguments:
         - *default* --
           the default value for this item
         - *doc* (*str*, default: *None*) --
           an optional description of this item
         - *namespace* (*dict*, default: *None*) --
           an optional namespace dictionary which will be made available
           to the Python code in the external configuration file
           during loading

        A configuration item is the object that stores the actual
        configuration settings.  Each item has a default value, a current
        value, an optional description, and an optional namespace.

        This class performs the checking of configuration values assigned
        to it during loading of the configuration file.  The default
        behavior of this class is to only accept values of the same Python
        type as the item's default value.  So, if the default value is a
        string, then the value assigned in the configuration file must
        also be a string.  Otherwise an exception will be raised and
        loading will fail.

        Developers who want other kinds of value checking should override
        the :meth:`Item.validate` method of this class.

    """


    def __init__(self, default, doc=None, namespace=None):
        self._default = default
        self._value = default
        self.doc = doc
        self._namespace = namespace

    def get_default(self):
        return self._default

    def get_value(self):
        return self._value

    def set_value(self, value):
        self.validate(value)
        self._value = value

    def validate(self, value):
        """
            Determine whether the given *value* is valid.

            This method performs validity checking of the configuration
            value assigned to this item during loading of the external
            configuration file.  If the default behavior is to raise a
            *TypeError* if the type of the assigned value is not the same
            as the type of the default value.

        """
        if not isinstance(value, type(self._default)):
            raise TypeError("Invalid type %s, expecting %s."
                            % (type(value), type(self._default)))

    value = property(get_value, set_value)
    default = property(get_default)

    def update_namespace(self, namespace):
        if self._namespace:
            namespace.update(self._namespace)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
SR back-end package for DNS and Natlink
============================================================================

"""

import logging
import os
import platform
import struct
import sys

_log = logging.getLogger("engine.natlink")


#---------------------------------------------------------------------------

# Module level singleton instance of this engine implementation.
_engine = None


def is_engine_available(**kwargs):
    """
        Check whether Natlink is available.

        :param \\**kwargs: optional keyword arguments passed through to the
            engine for engine-specific configuration.
    """
    # pylint: disable=too-many-return-statements
    global _engine
    if _engine:
        return True

    platform_name = platform.system()
    if platform_name != 'Windows':
        _log.warning("%s is not supported by the Natlink engine backend",
                     platform_name)
        return False

    if struct.calcsize("P") == 8:  # 64-bit
        _log.warning("The python environment is 64-bit. Natlink requires a "
                     "32-bit python environment")
        return False

    # Attempt to import natlink.
    try:
        import natlink
    except ImportError as e:
        # Add Natlink's default 'core' directory path to sys.path if
        # necessary.
        coredir_path = r'C:\\NatLink\\NatLink\\MacroSystem\\core'
        pyd_filename = 'natlink.pyd'
        pyd_path = os.path.join(coredir_path, pyd_filename)
        import_failure = True
        if os.path.isdir(coredir_path):
            if not os.path.isfile(pyd_path):
                _log.warning("Requested engine 'natlink' is not available: "
                             "The %r file is missing from Natlink's core "
                             "directory", pyd_filename)
                return False

            # Add the core directory to the path and try importing again.
            sys.path.append(coredir_path)
            try:
                import natlink
                import_failure = False
            except ImportError:
                pass
        if import_failure:
            _log.warning("Requested engine 'natlink' is not available: "
                         "Natlink is not installed: %s", e)
            return False
    except Exception as e:
        _log.exception("Exception during import of natlink package: "
                       "%s", e)
        return False

    try:
        if natlink.isNatSpeakRunning():
            return True
        else:
            _log.warning("Requested engine 'natlink' is not available: "
                         "Dragon NaturallySpeaking is not running")
            return False
    except Exception as e:
        _log.exception("Exception during natlink.isNatSpeakRunning(): "
                       "%s", e)
        return False


def get_engine(**kwargs):
    """
        Retrieve the Natlink back-end engine object.

        :param \\**kwargs: optional keyword arguments passed through to the
            engine for engine-specific configuration.
    """
    global _engine
    if not _engine:
        from .engine import NatlinkEngine
        _engine = NatlinkEngine(**kwargs)
    return _engine



# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
RPC utility functions
----------------------------------------------------------------------------
"""

import json
import requests


def send_rpc_request(url, method, params, id=0):
    """
    Utility function to send a JSON-RPC request to a server.

    This will raise an error if the request fails with an error or if
    the server is unreachable.

    :param url: the URL to send the JSON-RPC request to.
    :param method: name of the RPC method to call.
    :param params: parameters of the RPC method to call.
    :param id: ID of the JSON-RPC request (default: 0).
    :type url: str
    :type method: str
    :type params: list
    :type id: int
    :returns: JSON-RPC response
    :rtype: dict
    :raises: RuntimeError
    """
    headers = {'content-type': 'application/json'}
    payload = {
        "method": method,
        "params": params,
        "jsonrpc": "2.0",
        "id": id,
    }
    response = requests.post(
        url, data=json.dumps(payload), headers=headers
    ).json()

    # Handle an error in the response.
    error_info = response.get("error")
    if error_info:
        raise RuntimeError(
            "error calling RPC server method %s: %d, %s" %
            (method, error_info["code"], error_info["message"])
        )

    # Otherwise return the response.
    return response



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import sys
import os

# Windows
if sys.platform.startswith("win"):
    from .win32_monitor import Win32Monitor as Monitor

# Mac OS
elif sys.platform == "darwin":
    from .darwin_monitor import DarwinMonitor as Monitor

# Linux/X11
elif os.environ.get("XDG_SESSION_TYPE") == "x11":
    from .x11_monitor import X11Monitor as Monitor

# Unsupported
else:
    from .base_monitor import FakeMonitor as Monitor


class MonitorList(object):
    """
    Special read-only, self-updating monitors list class.

    Supports indexing and iteration.
    """

    def __init__(self):
        self._list = None  # lazily initialised

    def _update(self):
        self._list = Monitor.get_all_monitors()

    def __getitem__(self, index):
        self._update()
        return self._list[index]

    def __iter__(self):
        self._update()
        return iter(self._list)


#: :class:`MonitorsList` instance
monitors = MonitorList()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Recognition observer base class
============================================================================

"""

import logging

try:
    from inspect import getfullargspec as getargspec
except ImportError:
    # Fallback on the deprecated function.
    from inspect import getargspec

#---------------------------------------------------------------------------


class RecObsManagerBase(object):

    _log = logging.getLogger("engine.recobs")

    def __init__(self, engine):
        self._engine = engine
        self._enabled = True
        self._observers = []
        self._observer_ids = set()

    def enable(self):
        if not self._enabled:
            self._enabled = True
            self._activate()

    def disable(self):
        if self._enabled:
            self._enabled = False
            self._deactivate()

    def register(self, observer):
        if id(observer) in self._observer_ids:
            return
        elif not self._observers:
            self._activate()
        self._observers.append(observer)
        self._observer_ids.add(id(observer))

    def unregister(self, observer):
        try:
            self._observers.remove(observer)
            self._observer_ids.remove(id(observer))
        except ValueError:
            pass
        else:
            if not self._observers:
                self._deactivate()

    def _process_observer_callbacks(self, cb_name, required_names,
                                    **kwargs):
        for observer in self._observers:
            func = getattr(observer, cb_name, None)
            if not func:
                continue

            # If the callback function takes keyword arguments, only send
            # those that it accepts. Always pass required names.
            func_kwargs = kwargs
            if func_kwargs:
                argspec = getargspec(func)
                arg_names, kwargs_names = argspec[0], argspec[2]
                if not kwargs_names:
                    func_kwargs = {k: v for (k, v) in func_kwargs.items()
                                   if k in arg_names or k in required_names}

            # Call the callback function, catching and logging exceptions.
            try:
                func(**func_kwargs)
            except Exception as e:
                self._log.exception("Exception during %s()"
                                    " method of recognition observer %s: %s"
                                    % (cb_name, observer, e))

    def notify_begin(self):
        self._process_observer_callbacks("on_begin", [])

    def notify_recognition(self, words, rule, node, results):
        self._process_observer_callbacks("on_recognition", ["words"],
                                         words=words, rule=rule, node=node,
                                         results=results)
        self.notify_end(results)

    def notify_failure(self, results):
        self._process_observer_callbacks("on_failure", [], results=results)
        self.notify_end(results)

    def notify_end(self, results):
        self._process_observer_callbacks("on_end", [], results=results)

    def notify_post_recognition(self, words, rule, node, results):
        self._process_observer_callbacks("on_post_recognition", ["words"],
                                         words=words, rule=rule, node=node,
                                         results=results)

    def _activate(self):
        raise NotImplementedError(str(self))

    def _deactivate(self):
        raise NotImplementedError(str(self))



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
RunCommand action
============================================================================

The :class:`RunCommand` action takes a command-line program to run including
any required arguments. On execution, the program will be started as a
subprocess.

Processing will occur asynchronously by default. Commands running
asynchronously should not normally prevent the Python process from exiting.

It may sometimes be necessary to use a list for the action's *command*
argument instead of a string. This is because some command-line shells may
not work 100% correctly with Python's built-in :meth:`shlex.split` function.

This action should work on Windows and other platforms.

Example using the ping command::

    from dragonfly import RunCommand

    # Ping localhost for 4 seconds.
    RunCommand('ping -w 4 localhost').execute()


Example using a command list instead of a string::

    from dragonfly import RunCommand

    # Ping localhost for 4 seconds.
    RunCommand(['ping', '-w', '4', 'localhost']).execute()


Example using the optional function parameter::

    from __future__ import print_function
    from locale import getpreferredencoding
    from six import binary_type
    from dragonfly import RunCommand

    def func(proc):
        # Read lines from the process.
        encoding = getpreferredencoding()
        for line in iter(proc.stdout.readline, b''):
            if isinstance(line, binary_type):
                line = line.decode(encoding)
            print(line, end='')

    RunCommand('ping -w 4 localhost', func).execute()


Example using the optional synchronous parameter::

    from dragonfly import RunCommand

    RunCommand('ping -w 4 localhost', synchronous=True).execute()


Example using the optional hide_window parameter::

    from dragonfly import RunCommand

    # Use hide_window=False for running GUI applications via RunCommand.
    RunCommand('notepad.exe', hide_window=False).execute()


Example using the subprocess's :class:`Popen` object::

    from dragonfly import RunCommand

    # Initialise and execute a command asynchronously.
    cmd = RunCommand('ping -w 4 localhost')
    cmd.execute()

    # Wait until the subprocess finishes.
    cmd.process.wait()


Example using a subclass::

    from __future__ import print_function
    from locale import getpreferredencoding
    from six import binary_type
    from dragonfly import RunCommand

    class Ping(RunCommand):
        command = "ping -w 4 localhost"
        synchronous = True
        def process_command(self, proc):
            # Read lines from the process.
            encoding = getpreferredencoding()
            for line in iter(proc.stdout.readline, b''):
                if isinstance(line, binary_type):
                    line = line.decode(encoding)
                print(line, end='')

    Ping().execute()


Class reference
----------------------------------------------------------------------------

"""

from __future__ import print_function

import locale
import os
import shlex
import subprocess
import threading

from six import string_types, binary_type

from .action_base import ActionBase

# --------------------------------------------------------------------------


class RunCommand(ActionBase):
    """
        Start an application from the command-line.

        This class is similar to the :class:`StartApp` class, but is
        designed for running command-line applications and optionally
        processing subprocesses.

    """
    command = None
    synchronous = False

    def __init__(self, command=None, process_command=None,
                 synchronous=False, hide_window=True):
        """
            Constructor arguments:
             - *command* (str or list) -- the command to run when this
               action is executed. It will be parsed by :meth:`shlex.split`
               if it is a string and passed directly to ``subprocess.Popen``
               if it is a list. Command arguments can be included.
             - *process_command* (callable) -- optional callable to invoke
               with the :class:`Popen` object after successfully starting
               the subprocess. Using this argument overrides the
               :meth:`process_command` method.
             - *synchronous* (bool, default *False*) -- whether to wait
               until :meth:`process_command` has finished executing before
               continuing.
             - *hide_window* (bool, default *True*) -- whether to hide the
               application window. Set to *False* if using this action with
               GUI programs. This argument only applies to Windows. It has
               no effect on other platforms.

        """
        ActionBase.__init__(self)
        self._proc = None

        # Complex handling of arguments because of clashing use of the names
        # at the class level: property & class-value.
        if command is not None:
            self.command = command
        command_types = (string_types, list)
        if not (self.command and isinstance(self.command, command_types)):
            raise TypeError("command must be a non-empty string or list, "
                            "not %s" % self.command)
        if synchronous is not False:
            self.synchronous = synchronous

        if not (process_command is None or callable(process_command)):
            raise TypeError("process_command must be a callable object or "
                            "None")

        self._process_command = process_command
        self._hide_window = hide_window

        # Set the string used for representing actions.
        if isinstance(self.command, list):
            self._str = "'%s'" % " ".join(self.command)
        else:
            self._str = "'%s'" % self.command

    @property
    def process(self):
        """
            The :class:`Popen` object for the current subprocess if one has
            been started, otherwise ``None``.
        """
        return self._proc

    # pylint: disable=no-self-use
    def process_command(self, proc):
        """
            Method to override for custom handling of the command's
            :class:`Popen` object.

            By default this method prints lines from the subprocess until it
            exits.
        """
        encoding = locale.getpreferredencoding()
        for line in iter(proc.stdout.readline, b''):
            if isinstance(line, binary_type):
                line = line.decode(encoding)

            print(line, end='')

    def _execute(self, data=None):
        self._log.info("Executing: %s", self.command)

        # Suppress showing the new CMD.exe window on Windows.
        startupinfo = None
        if os.name == 'nt' and self._hide_window:
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW

        # Pre-process self.command before passing it to subprocess.Popen.
        command = self.command
        if isinstance(command, string_types):
            # Split command strings using shlex before passing it to Popen.
            # Use POSIX mode only if on a POSIX platform.
            command = shlex.split(command, posix=os.name == "posix")
        try:
            self._proc = subprocess.Popen(command,
                                          stdout=subprocess.PIPE,
                                          stderr=subprocess.STDOUT,
                                          stdin=subprocess.PIPE,
                                          startupinfo=startupinfo)
        except Exception as e:
            self._log.exception("Exception from starting subprocess %s: "
                                "%s", self._str, e)
            return False

        # Call process_command either synchronously or asynchronously.
        def call():
            try:
                if self._process_command:
                    process_func = self._process_command
                else:
                    process_func = self.process_command
                process_func(self._proc)
                return_code = self._proc.wait()
                if return_code != 0:
                    self._log.error("Command %s failed with return code "
                                    "%d", self._str, return_code)
                    return False
                return True
            except Exception as e:
                self._log.exception("Exception processing command %s: %s",
                                    self._str, e)
                return False
            finally:
                self._proc = None

        if self.synchronous:
            return call()

        # Execute in a new daemonized thread so that the command cannot
        # stop the SR engine from exiting.
        thread = threading.Thread(target=call)
        thread.setDaemon(True)
        thread.start()
        return True



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Dictation formatting for Natlink
============================================================================

This module implements dictation formatting for the Natlink and Dragon
NaturallySpeaking engine.

"""

# pylint: disable=attribute-defined-outside-init,redefined-builtin

from locale import getpreferredencoding
import logging
import re

# Attempt to import Natlink.  If this fails, Natlink is probably
#  not available and the dictation container implemented here
#  cannot be used.  However, we don't raise an exception because
#  this file should still be importable for documentation purposes.

try:
    import natlink
except ImportError:
    natlink = None

from six import binary_type, string_types, PY2

import dragonfly.engines


#===========================================================================

class FlagContainer(object):
    """ Container for a predefined set of boolean flags. """

    flag_names = ()

    def __init__(self, *names):
        self._flags_true = set()
        for name in names:
            setattr(self, name, True)

    def flags_string(self):
        flags_true_names = []
        for flag in self.flag_names:
            if flag in self._flags_true:
                flags_true_names.append(flag)
        return u", ".join(flags_true_names)

    def __unicode__(self):
        return u"%s(%s)" % (self.__class__.__name__, self.flags_string())

    def __repr__(self):
        if PY2:
            return self.__unicode__().encode(getpreferredencoding())
        else:
            return self.__unicode__()

    def __getattr__(self, name):
        if name not in self.flag_names:
            raise AttributeError("Invalid flag name: %r" % (name,))
        return name in self._flags_true

    # pylint: disable=inconsistent-return-statements
    def __setattr__(self, name, value):
        if name.startswith("_"):
            return object.__setattr__(self, name, value)
        if name not in self.flag_names:
            raise AttributeError("Invalid flag name: received %r, expected"
                                 " one of %r" % (name, self.flag_names))
        if value:
            self._flags_true.add(name)
        else:
            self._flags_true.discard(name)

    def clone(self):
        clone = self.__class__()
        for name in self._flags_true:
            setattr(clone, name, True)
        return clone


#---------------------------------------------------------------------------

class WordFlags(FlagContainer):
    """
        Container for formatting flags associated with DNS words.

        The flags defined by this class are closely related to the
        formatting information provided by DNS.

    """

    # pylint: disable=line-too-long
    flag_names = (
        # Flags related to spacing
        "no_space_before",     # No space before this word (like right-paren)
        "no_space_after",      # No space after this word (like left-paren)
        "two_spaces_after",    # Two spaces after this word (like full-stop)
        "no_space_mode",       # Activate no-spacing mode (like No-Space-On)
        "reset_no_space",      # Reset spacing mode (like No-Space-Off)
        "no_space_reset",      # Don't reset spacing state (like Cap)
        "spacebar",            # Extra space after this word (like spacebar)
        "no_space_between",    # No spaces between adjacent words with this flag (like numbers)

        # Flags related to newlines
        "newline_after",       # One newline after this word (like New-Line)
        "two_newlines_after",  # Two newlines after this word (like New-Paragraph)

        # Flags related to capitalization
        "cap_next",            # Normally capitalize next word (eg full-stop)
        "cap_next_force",      # Always capitalize next word (eg Cap-Next)
        "lower_next",          # Lowercase next word (eg No-Caps-Next)
        "upper_next",          # Uppercase next word (eg All-Caps-Next)
        "cap_mode",            # Activate capitalization mode (like Caps-On)
        "lower_mode",          # Activate lowercase mode (like No-Caps-On)
        "upper_mode",          # Activate uppercase mode (like All-Caps-On)
        "reset_cap",           # Reset capitalization mode (like Caps-Off)
        "no_cap_reset",        # Don't reset the capitalization state (like left-paren)
        "no_title_cap",        # Don't capitalize in title (like and)

        # Miscellaneous flags
        "no_format",           # Don't apply formatting (like Cap)
        "not_after_period",    # Suppress this word after a word ending in period (like full-stop after etc.)
    )


#---------------------------------------------------------------------------

class StateFlags(FlagContainer):
    """
        Container for keeping state for inter-word formatting.

        The flags defined by this class are used by Dragonfly to store
        formatting state between words as a formatter consumes
        consecutive words.

    """

    # pylint: disable=line-too-long,too-many-instance-attributes
    flag_names = (
        # Flags related to spacing
        "no_space_before",       # No space before next word
        "two_spaces_before",     # Two spaces before next word
        "no_space_between",      # No space before next word if it has no_space_between flag
        "no_space_mode",         # No-spacing mode is active

        # Flags related to capitalization
        "cap_next",              # Normally capitalize next word
        "cap_next_force",        # Always capitalize next word
        "lower_next",            # Lowercase next word
        "upper_next",            # Uppercase next word
        "cap_mode",              # Capitalization mode is active
        "lower_mode",            # Lowercase mode is active
        "upper_mode",            # Uppercase mode is active

        # Miscellaneous flags
        "prev_ended_in_period",  # Previous word ended in period
    )


#===========================================================================

class Word(object):
    """ Word storing written and spoken forms with formatting flags. """

    def __init__(self, written, spoken, flags):
        self.written = written
        self.spoken = spoken
        self.flags = flags

    def __unicode__(self):
        info = [repr(self.written)]
        if self.spoken and self.spoken != self.written:
            info.append(repr(self.spoken))
        flags_string = self.flags.flags_string()
        if flags_string:
            info.append(flags_string)
        return u"%s(%s)" % (self.__class__.__name__, ", ".join(info))

    def __repr__(self):
        if PY2:
            return self.__unicode__().encode(getpreferredencoding())
        else:
            return self.__unicode__()


#===========================================================================

class WordParserBase(object):
    """ Base class for parsing dictation results into Word objects. """

    _log = logging.getLogger("dictation.word_parser")

    def parse_input(self, input):
        raise NotImplementedError


#---------------------------------------------------------------------------

class WordParserDns10(WordParserBase):
    """
        Dictation results parser for DNS v10 and lower.

        DNS v10 and lower provide word formatting information as a 32-bit
        value returned by `natlink.getWordInfo()`.

    """

    flags = [
        (0x00000001, None),  # Added by user
        (0x00000002, None),  # Internal use
        (0x00000004, None),  # Internal usu
        (0x00000008, None),  # Undeletable
        (0x00000010, "cap_next"),
        (0x00000020, "cap_next_force"),
        (0x00000040, "upper_next"),
        (0x00000080, "lower_next"),
        (0x00000100, "no_space_after"),
        (0x00000200, "two_spaces_after"),
        (0x00000400, "no_space_between"),
        (0x00000800, "cap_mode"),
        (0x00001000, "upper_mode"),
        (0x00002000, "lower_mode"),
        (0x00004000, "no_space_mode"),
        (0x00008000, "reset_no_space"),
        (0x00010000, None),  # Internal use
        (0x00020000, "not_after_period"),
        (0x00040000, "no_format"),
        (0x00080000, "no_space_reset"),
        (0x00100000, "no_cap_reset"),
        (0x00200000, "no_space_before"),
        (0x00400000, "reset_cap"),
        (0x00800000, "newline_after"),
        (0x01000000, "two_newlines_after"),
        (0x02000000, "no_title_cap"),
        (0x04000000, None),  # Internal use
        (0x08000000, "spacebar"),
        (0x10000000, None),  # Internal use
        (0x20000000, None),  # Internal use
        (0x40000000, None),  # Added by vocabulary builder
        (0x80000000, None),
    ]

    def create_word_flags(self, info):
        if info is None:
            # None indicates the word is not in DNS' vocabulary.
            return WordFlags()

        flag_names = []
        for bit, name in self.flags:
            if (info & bit) != 0:
                if name:
                    flag_names.append(name)
        return WordFlags(*flag_names)

    def parse_input(self, input):
        if not isinstance(input, string_types):
            raise TypeError("input must be a string, not {0!r}"
                            .format(input))

        if isinstance(input, binary_type):
            # DNS and Natlink provide recognized words as "Windows-1252"
            # encoded strings. Here we convert them to Unicode for internal
            # processing.
            input = input.decode("windows-1252")

        # The written and spoken forms of a word are separated by a "\"
        # character.
        index = input.rfind("\\")
        if index == -1:
            # Input doesn't contain a backslash, so written and spoken forms
            # are the same as the input.
            written = input
            spoken = input
        else:
            # Input contains one or more backslashes, so written and spoken
            # forms are separated by the last backslash.
            written = input[:index]
            spoken = input[index+1:]

        info = natlink.getWordInfo(input.encode("windows-1252"))
        word_flags = self.create_word_flags(info)

        word = Word(written, spoken, word_flags)
        self._log.debug("Parsed input %r -> %s", input, word)
#        print ("Parsed input {0!r} -> {1}".format(input, word))
        return word


#---------------------------------------------------------------------------

class WordParserDns11(WordParserBase):
    """
        Dictation results parser for DNS v11 and higher.

        DNS v11 and higher provide word formatting information ...

    """

    # pylint: disable=line-too-long
    property_map = {
        "new-line":         WordFlags("no_format", "no_space_after", "no_cap_reset", "newline_after"),
        "new-paragraph":    WordFlags("no_format", "no_space_after", "cap_next", "two_newlines_after"),
        "no-space":         WordFlags("no_format", "no_cap_reset", "no_space_after"),
        "no-space-on":      WordFlags("no_format", "no_cap_reset", "no_space_mode"),
        "no-space-off":     WordFlags("no_format", "no_cap_reset", "reset_no_space"),

        "cap":              WordFlags("no_format", "no_space_reset", "cap_next_force"),
        "caps-on":          WordFlags("no_format", "no_space_reset", "reset_cap", "cap_mode"),
        "caps-off":         WordFlags("no_format", "no_space_reset", "reset_cap"),
        "all-caps":         WordFlags("no_format", "no_space_reset", "upper_next"),
        "all-caps-on":      WordFlags("no_format", "no_space_reset", "reset_cap", "upper_mode"),
        "all-caps-off":     WordFlags("no_format", "no_space_reset", "reset_cap"),
        "no-caps":          WordFlags("no_format", "no_space_reset", "lower_next"),
        "no-caps-on":       WordFlags("no_format", "no_space_reset", "reset_cap", "lower_mode"),
        "no-caps-off":      WordFlags("no_format", "no_space_reset", "reset_cap"),

        "space-bar":        WordFlags("no_format", "spacebar", "no_space_after", "no_cap_reset", "no_space_before"),
        "spelling-cap":     WordFlags("no_format", "no_space_reset", "cap_next_force"),
        "letter":           WordFlags("no_space_between"),
        "uppercase-letter": WordFlags("no_space_between"),
        "numeral":          WordFlags("no_space_after"),

        "period":           WordFlags("two_spaces_after", "cap_next", "no_space_before", "not_after_period"),
        "question-mark":    WordFlags("two_spaces_after", "cap_next", "no_space_before"),
        "exclamation-mark": WordFlags("two_spaces_after", "cap_next", "no_space_before"),
        "point":            WordFlags("no_space_after", "no_space_between", "no_space_before"),
        "dot":              WordFlags("no_space_after", "no_space_between", "no_space_before"),
        "ellipsis":         WordFlags("no_space_before", "not_after_period"),
        "comma":            WordFlags("no_space_before"),
        "hyphen":           WordFlags("no_space_before", "no_space_after"),
        "dash":             WordFlags("no_space_before", "no_space_after"),
        "at-sign":          WordFlags("no_space_before", "no_space_after"),
        "colon":            WordFlags("no_space_before"),
        "semicolon":        WordFlags("no_space_before"),
        "apostrophe-ess":   WordFlags("no_space_before"),
        "left-*":           WordFlags("no_cap_reset", "no_space_after"),
        "right-*":          WordFlags("no_cap_reset", "no_space_before", "no_space_reset"),
        "open paren":       WordFlags("no_space_after"),
        "close paren":      WordFlags("no_space_before"),
        "slash":            WordFlags("no_space_after", "no_space_before"),

        # below are two examples of Dragon custom vocabulary with formatting
        # these would have to be added to the Dragon vocabulary for users to use them
        # "len":              WordFlags("no_space_after"), # shorter name for (
        # "ren":              WordFlags("no_space_before"), # shorter name for )
    }

    def create_word_flags(self, property):
        if not property:
            # None indicates the word is not in DNS' vocabulary.
            flags = WordFlags()
        elif property in self.property_map:
            flags = self.property_map[property].clone()
        elif property.startswith("left-"):
            flags = self.property_map["left-*"].clone()
        elif property.startswith("right-"):
            flags = self.property_map["right-*"].clone()
        else:
            self._log.warning("Unknown word property: %r", property)
            flags = WordFlags()
        return flags

    def parse_input(self, input):
        if not isinstance(input, string_types):
            raise TypeError("input must be a string, not {0!r}"
                            .format(input))

        if isinstance(input, binary_type):
            # DNS and Natlink provide recognized words as "Windows-1252"
            # encoded strings. Here we convert them to Unicode for internal
            # processing.
            input = input.decode("windows-1252")

        parts = input.split("\\")
        if len(parts) == 1:
            # Word doesn't have "written\property\spoken" form, so
            # written and spoken forms are equal to input and there are
            # no formatting flags.
            written = input
            spoken = input
            property = None
        elif len(parts) == 2:
            if parts[1] == "letter":
                # Format: "X \ letter"
                written = spoken = parts[0]
                property = "letter"
            else:
                # Format: "written \ spoken"
                written, spoken = parts
                property = None
        elif len(parts) == 3:
            # Format: "written \ property \ spoken"
            written, property, spoken = parts
        else:  # len(parts) > 3:
            # Format: "wri\tt\en \ property \ spoken"
            written = "\\".join(parts[:-2])
            property = parts[-2]
            spoken = parts[-1]

        word_flags = self.create_word_flags(property)

        word = Word(written, spoken, word_flags)
        self._log.debug("Parsed input %r -> %s", input, word)
#        print ("Parsed input {0!r} -> {1}".format(input, word))
        return word


#===========================================================================

class WordParserFactory(object):
    """
        Detects appropriate WordParser class and creates objects

        This class calls natlink.getWordInfo() for special words and
        interprets the result to determine whether words should be parsed
        using natlink.getWordInfo(), e.g. for DNS v10 and lower, or using
        "in-line word properties", e.g. for DNS v11 and higher.

        This class performs detection only once per instance and caches
        the detection result for reuse.

    """

    _log = logging.getLogger("dictation.word_parser_factory")

    # The following dictionary determines which words are used for which
    # language to detect whether natlink.getWordInfo() returns useful info.
    words_with_info = {
        "en": ".\\dot",
        "nl": ".\\dot",
    }

    def __init__(self):
        self.parser_class = None

    def detect_parser_class(self):
        engine = dragonfly.engines.get_engine()
        word = self.words_with_info.get(engine.language, ".\\dot")
        if PY2:
            info = natlink.getWordInfo(word.encode("windows-1252"))
        else:
            info = natlink.getWordInfo(word)
        if info is None:
            parser_class = WordParserDns11
        else:
            parser_class = WordParserDns10

        parser_class_string = parser_class.__name__
        if info is None:  info_string = "None"
        else:             info_string = "0x" + format(info, "08x")
        self._log.info("Selected word parser class %s because"
                       " natlink.getWordInfo(%r) returned %s",
                       parser_class_string, word, info_string)

        return parser_class

    def get_parser(self):
        """ Create an instance of the detective parser class. """
        if not self.parser_class:
            self.parser_class = self.detect_parser_class()
        return self.parser_class()


#===========================================================================

class WordFormatter(object):

    _log = logging.getLogger("dictation.formatter")

    parser_factory = WordParserFactory()

    def __init__(self, state=None, parser=None,
                 two_spaces_after_period=False):
        if state:   self.state = state
        else:       self.state = StateFlags("no_space_before")
        if parser:  self.parser = parser
        else:       self.parser = self.parser_factory.get_parser()
        self.two_spaces_after_period = two_spaces_after_period

    #-----------------------------------------------------------------------
    # Main formatting methods.
    # pylint: disable=bad-indentation,too-many-branches,line-too-long

    def format_dictation(self, input_words):
        if isinstance(input_words, string_types):
            raise ValueError("Argument input_words must be a sequence of"
                             " words, but received a single string: {0!r}"
                             .format(input_words))

        formatted_words = []
        for input_word in input_words:
            word = self.parser.parse_input(input_word)
            formatted_words.append(self.apply_formatting(word))
            new_state = self.update_state(word)
            self._log.debug("Processing %s, formatted output: %r,"
                            " %s -> %s",
                            word, formatted_words[-1],  self.state,
                            new_state)
            self.state = new_state
        return u"".join(formatted_words)

    def apply_formatting(self, word):
        state = self.state

        # Determine prefix.
        if   word.flags.no_format:             prefix = ""
        elif state.no_space_mode:              prefix = ""
        elif word.flags.no_space_before:       prefix = ""
        elif state.no_space_before:            prefix = ""
        elif state.no_space_between and word.flags.no_space_between:
                                               prefix = ""
        elif state.two_spaces_before:
            if self.two_spaces_after_period:   prefix = "  "
            else:                              prefix = " "
        else:                                  prefix = " "

        # Determine formatted written form.
        if   word.flags.no_format:         written = word.written
        elif state.cap_mode and not word.flags.no_title_cap:
                                           written = self.capitalize_all(word.written)
        elif state.upper_mode:             written = self.upper_all(word.written)
        elif state.lower_mode:             written = self.lower_all(word.written)
        elif state.cap_next:               written = self.capitalize_first(word.written)
        elif state.cap_next_force:         written = self.capitalize_first(word.written)
        elif state.upper_next:             written = self.upper_first(word.written)
        elif state.lower_next:             written = self.lower_first(word.written)
        else:                              written = word.written

        # Remove first period character if needed.
        if (state.prev_ended_in_period
                and word.flags.not_after_period
                and written.startswith(".")):
            written = written[1:]

        # Determine suffix.
        if   word.flags.newline_after:       suffix = "\n"
        elif word.flags.two_newlines_after:  suffix = "\n\n"
        elif word.flags.spacebar:            suffix = " "
        else:                                suffix = ""

        return prefix + written + suffix

    def update_state(self, word_object):
        word = word_object.flags
        prev = self.state
        state = StateFlags()

        # Update capitalization state.
        state.cap_next_force = word.cap_next_force or (word.no_cap_reset and prev.cap_next_force)
        state.cap_next       = word.cap_next or word.cap_mode or (word.no_cap_reset and prev.cap_next)
        state.upper_next     = word.upper_next     or (word.no_cap_reset and prev.upper_next)
        state.lower_next     = word.lower_next     or (word.no_cap_reset and prev.lower_next)
        state.cap_mode       = word.cap_mode       or (prev.cap_mode   and not word.reset_cap)
        state.upper_mode     = word.upper_mode     or (prev.upper_mode and not word.reset_cap)
        state.lower_mode     = word.lower_mode     or (prev.lower_mode and not word.reset_cap)

        # Update spacing state.
        state.no_space_before   = word.no_space_after   or (prev.no_space_before and word.no_space_reset and word.no_format)
        state.two_spaces_before = word.two_spaces_after or (prev.two_spaces_before and word.no_space_reset and word.no_format)
        state.no_space_between  = word.no_space_between
        state.no_space_mode     = word.no_space_mode    or (prev.no_space_mode and not word.reset_no_space)

        # Record whether this word ended in a period.
        state.prev_ended_in_period = word_object.written.endswith(".")

        # Return newly created state.
        return state

    #-----------------------------------------------------------------------
    # String formatting utility methods.
    # pylint: disable=no-self-use

    def capitalize_first(self, written):
        return written.capitalize()

    def _capitalize_all_function(self, match):
        return match.group(1) + match.group(2).upper()

    def capitalize_all(self, written):
        return re.sub(r"(^|\s)(\S)", self._capitalize_all_function, written)

    def upper_first(self, written):
        parts = written.split(" ", 1)
        parts[0] = parts[0].upper()
        return " ".join(parts)

    def lower_first(self, written):
        parts = written.split(" ", 1)
        parts[0] = parts[0].lower()
        return " ".join(parts)

    def upper_all(self, written):
        return written.upper()

    def lower_all(self, written):
        return written.lower()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
WaitWindow action
============================================================================

"""


import time

from dragonfly.actions.action_base import ActionBase, ActionError
from ..windows import Window


#---------------------------------------------------------------------------

class WaitWindow(ActionBase):
    """
        Wait for a specific window context action.

        Constructor arguments:
         - *title* (*str*) --
           part of the window title: not case sensitive
         - *executable* (*str*) --
           part of the file name of the executable; not case sensitive
         - *timeout* (*int* or *float*) --
           the maximum number of seconds to wait for the correct
           context, after which an :class:`ActionError` will
           be raised.

        When this action is executed, it waits until the correct window
        context is present.  This window context is specified by the
        desired window title of the foreground window and/or the
        executable name of the foreground application.  These are
        specified using the constructor arguments listed above.  The
        substring search used is *not* case sensitive.

        If the correct window context is not found within *timeout*
        seconds, then this action will raise an :class:`ActionError` to
        indicate the timeout.

    """

    def __init__(self, title=None, executable=None,
                 timeout=15):
        self._match_functions = []
        string = []

        if title is not None:
            self._title = title.lower()
            self._match_functions.append("_match_title")
            string.append("title=%r" % self._title)
        else:
            self._title = None

        if executable is not None:
            self._executable = executable.lower()
            self._match_functions.append("_match_executable")
            string.append("executable=%r" % self._executable)
        else:
            self._executable = None

        self._timeout = timeout

        ActionBase.__init__(self)
        self._str = ", ".join(string)

    def _execute(self, data=None):
        self._log.debug("Waiting for window context: %s", self)
        start_time = time.time()
        while 1:
            foreground = Window.get_foreground()
            mismatch = False
            for match_name in self._match_functions:
                match_func = getattr(self, match_name)
                if not match_func(foreground):
                    mismatch = True
                    break
            if not mismatch:
                return
            if time.time() - start_time > self._timeout:
                raise ActionError("Timeout while waiting for window context: %s" % self)

    def _match_title(self, foreground):
        if self._title is None:
            return True
        current_title = foreground.title.lower()
        if current_title.find(self._title) != -1:
            return True
        return False

    def _match_executable(self, foreground):
        if self._executable is None:
            return True
        current_executable = foreground.executable.lower()
        if current_executable.find(self._executable) != -1:
            return True
        return False



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Multiplexing interface to the Natlink timer
============================================================================

"""

from ..base import TimerManagerBase


#---------------------------------------------------------------------------

class NatlinkTimerManager(TimerManagerBase):
    """
    Timer manager for the Natlink engine.

    This class utilises :meth:`natlink.setTimerCallback` to ensure that
    timer functions are called on-time regardless of Dragon's current
    status.

    Python code run outside of timer functions will be blocked when
    ``natlink`` functions are executing. This is a
    `limitation with Python threads
    <https://docs.python.org/3/c-api/init.html#thread-state-and-the-global-interpreter-lock>`__.

    :meth:`engine.connect` must be called before using timers with this
    manager.

    **Note**: long-running timers will block dragonfly from processing
    what was said, so be careful with how you use them!
    """

    def __init__(self, interval, engine):
        """
        """
        TimerManagerBase.__init__(self, interval, engine)

    def _activate_main_callback(self, callback, sec):
        self.engine.natlink.setTimerCallback(callback, int(sec * 1000))

    def _deactivate_main_callback(self):
        self.engine.natlink.setTimerCallback(None, 0)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
CompoundRule class
============================================================================

The CompoundRule class is designed to make it very easy to create a rule
based on a single compound spec.

This rule class has the following parameters to customize its behavior:

 - *spec* -- compound specification for the rule's root element
 - *extras* -- extras elements referenced from the compound spec
 - *defaults* -- default values for the extras
 - *exported* -- whether the rule is exported
 - *context* -- context in which the rule will be active

Each of these parameters can be passed as a (keyword) arguments to the
constructor, or defined as a class attribute in a derived class.


Example usage
............................................................................

The CompoundRule class can be used to define a voice-command as follows::

    class ExampleRule(CompoundRule):

        spec = "I want to eat <food>"
        extras = [Choice("food", {
                                  "(an | a juicy) apple": "good",
                                  "a [greasy] hamburger": "bad",
                                 }
                        )
                 ]

        def _process_recognition(self, node, extras):
            good_or_bad = extras["food"]
            print "That is a %s idea!" % good_or_bad

    rule = ExampleRule()
    grammar.add_rule(rule)


Class reference
............................................................................

"""
from six import string_types

from .rule_base  import Rule
from .elements   import ElementBase, Compound


#---------------------------------------------------------------------------

class CompoundRule(Rule):
    """
        Rule class based on the compound element.

        Constructor arguments:
         - *name* (*str*) -- the rule's name
         - *spec* (*str*) -- compound specification for the rule's
           root element
         - *extras* (sequence) -- extras elements referenced from the
           compound spec
         - *defaults* (*dict*) -- default values for the extras
         - *exported* (boolean) -- whether the rule is exported
         - *context* (*Context*) -- context in which the rule will be active

    """

    _name    = None
    spec     = None
    extras   = ()
    defaults = ()
    context  = None
    _default_exported = True

    #-----------------------------------------------------------------------

    def __init__(self, name=None, spec=None, extras=None,
                 defaults=None, exported=None, context=None):
        if name     is None: name     = self._name or self.__class__.__name__
        if spec     is None: spec     = self.spec
        if extras   is None: extras   = self.extras
        if defaults is None: defaults = self.defaults
        if context  is None: context  = self.context

        # Complex handling of exported, because of clashing use of the
        #  exported name at the class level: property & class-value.
        if exported is not None:
            pass
        elif (hasattr(self.__class__, "exported")
              and not isinstance(self.__class__.exported, property)):
            exported = self.__class__.exported
        else:
            exported = self._default_exported

        assert isinstance(name, string_types)
        assert isinstance(spec, string_types)
        assert isinstance(extras, (list, tuple))
        for item in extras:
            assert isinstance(item, ElementBase)
        assert exported in (True, False)

        self._name     = name
        self._spec     = spec
        self.spec = spec
        self._extras   = dict((element.name, element) for element in extras)
        self._defaults = dict(defaults)

        child = Compound(spec, extras=self._extras)
        Rule.__init__(self, name, child, exported=exported, context=context)

    #-----------------------------------------------------------------------

    def process_recognition(self, node):
        """
            Process a recognition of this rule.

            This method is called by the containing Grammar when this
            rule is recognized.  This method collects information about
            the recognition and then calls *self._process_recognition*.

            - *node* -- The root node of the recognition parse tree.
        """
        # Prepare *extras* dict for passing to _process_recognition().
        extras = {
            "_grammar":  self.grammar,
            "_rule":     self,
            "_node":     node,
        }
        extras.update(self._defaults)
        for name, element in self._extras.items():
            extra_node = node.get_child_by_name(name, shallow=True)
            if extra_node:
                extras[name] = extra_node.value()
            elif element.has_default():
                extras[name] = element.default

        # Call the method to do the actual processing.
        self._process_recognition(node, extras)

    def _process_recognition(self, node, extras):
        """
            Default recognition processing.

            This is the method which should be overridden in most cases
            to provide derived classes with custom recognition
            processing functionality.

            This default processing method does nothing.

            - *node* -- The root node of the recognition parse tree.
            - *extras* -- A dictionary of all elements from the
              extras list contained within this recognition.
              Maps element name -> element value.
        """



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Test cases for the Natlink timer
============================================================================

"""

import unittest


#---------------------------------------------------------------------------

class NatlinkTimerManagerTest(unittest.TestCase):
    """ NatlinkTimerManager tests. """

    def DISABLED_test_natlink_timer_manager(self):
        return

        import dragonfly.engines.backend_natlink as backend
        print(backend.is_engine_available())
        engine = backend.get_engine()
        print(engine)
        engine.connect()
        try:
            print("starting timer...")
            def callback():
                engine._log.error("timer callback")
                print("timer callback")
            timer = engine.create_timer(callback, 1)
            print("timer:", timer)

            print("starting Luke...")
            import sys
            import time
            import win32gui
            timeout = time.time() + 3
            while time.time() < timeout:
                print("Luke")
                sys.stdout.flush()
                if win32gui.PumpWaitingMessages():
                    raise RuntimeError("We got an unexpected WM_QUIT message!")
                time.sleep(1)

        finally:
            engine.disconnect()

#        self.engine.natlink.setTimerCallback(callback, int(sec * 1000))


#---------------------------------------------------------------------------

class NatlinkTimerTest(object):#unittest.TestCase):
    """ Natlink timer tests. """

    def DISABLED_test_natlink_timer(self):
        return

        callback_occurred = False
        def callback():
            callback_occurred = True
        ticks = 200

        import time
        import natlink
        natlink.natConnect()
        try:
            natlink.setTimerCallback(callback, ticks)
#            time.sleep(1)
            import sys
            import time
            import win32gui
            timeout = time.time() + 3
            while time.time() < timeout:
                print("Luke")
                sys.stdout.flush()
                if win32gui.PumpWaitingMessages():
                    raise RuntimeError("We got an unexpected WM_QUIT message!")
                time.sleep(1)
            natlink.setTimerCallback(None, 0)
            print("callback occurred:", callback_occurred)
        finally:
            natlink.natDisconnect()


#---------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



"""
Multiple dictation constructs
===============================================================================

This file is a showcase investigating the use and functionality of multiple
dictation elements within Dragonfly speech recognition grammars.

The first part of this file (i.e. the module's doc string) contains a 
description of the functionality being investigated along with test code 
and actual output in doctest format. This allows the reader to see what 
really would happen, without needing to load the file into a speech 
recognition engine and put effort into speaking all the showcased 
commands.

The test code below makes use of Dragonfly's built-in element testing tool.
When run, it will connect to the speech recognition engine, load the element
being tested, mimic recognitions, and process the recognized value.


Multiple consecutive dictation elements
-------------------------------------------------------------------------------

>>> tester = ElementTester(RuleRef(ConsecutiveDictationRule()))
>>> print(tester.recognize("consecutive Alice Bob Charlie"))
Recognition: "consecutive Alice Bob Charlie"
Word and rule pairs: ("1000000" is "dgndictation")
 - consecutive (1)
 - Alice (1000000)
 - Bob (1000000)
 - Charlie (1000000)
Extras:
 - dictation1: Alice
 - dictation2: Bob
 - dictation3: Charlie

>>> print(tester.recognize("consecutive Alice Bob"))
RecognitionFailure


Mixed literal and dictation elements
-------------------------------------------------------------------------------

Here we will investigate mixed, i.e. interspersed, fixed literal command
words and dynamic dictation elements. We will use the "MixedDictationRule"
class which has a spec of
"mixed [<dictation1>] <dictation2> command <dictation3>".

Note that "<dictation1>" was made optional instead of "<dictation2>" 
because otherwise the first dictation elements would always gobble up 
all dictated words. There would (by definition) be no way to distinguish 
which words correspond with which dictation elements. Such consecutive 
dictation elements should for that reason be avoided in real command 
grammars. The way the spec is defined now, adds some interesting 
dynamics, because of the order in which they dictation elements parse 
the recognized words. However, do note that that order is well defined 
but arbitrarily chosen. 

>>> tester = ElementTester(RuleRef(MixedDictationRule()))
>>> print(tester.recognize("mixed Alice Bob command Charlie"))
Recognition: "mixed Alice Bob command Charlie"
Word and rule pairs: ("1000000" is "dgndictation")
 - mixed (1)
 - Alice (1000000)
 - Bob (1000000)
 - command (1)
 - Charlie (1000000)
Extras:
 - dictation1: Alice
 - dictation2: Bob
 - dictation3: Charlie

>>> print(tester.recognize("mixed Alice command Charlie"))
Recognition: "mixed Alice command Charlie"
Word and rule pairs: ("1000000" is "dgndictation")
 - mixed (1)
 - Alice (1000000)
 - command (1)
 - Charlie (1000000)
Extras:
 - dictation2: Alice
 - dictation3: Charlie

>>> print(tester.recognize("mixed Alice Bob command"))
RecognitionFailure

>>> print(tester.recognize("mixed command Charlie"))
RecognitionFailure


Repetition of dictation elements
-------------------------------------------------------------------------------

Now let's take a look at repetition of dictation elements. For this
we will use the "RepeatedDictationRule" class, which defines its spec
as a repetition of "command <dictation>". I.e. "command Alice" will
match, and "command Alice command Bob" will also match.

Note that this rule is inherently ambiguous, given the lack of a
clear definition of grouping or precedence rules for fixed literal
words in commands, and dynamic dictation elements. As an example,
"command Alice command Bob" could either match 2 repetitions with
"Alice" and "Bob" as dictation values, or a single repetition with
"Alice command Bob" as its only dictation value. The tests below
the show which of these actually occurs.

>>> tester = ElementTester(RuleRef(RepeatedDictationRule()))
>>> print(tester.recognize("command Alice"))
Recognition: "command Alice"
Word and rule pairs: ("1000000" is "dgndictation")
 - command (1)
 - Alice (1000000)
Extras:
 - repetition: [[u'command', NatlinkDictationContainer(Alice)]]

>>> print(tester.recognize("command Alice command Bob"))
Recognition: "command Alice command Bob"
Word and rule pairs: ("1000000" is "dgndictation")
 - command (1)
 - Alice (1000000)
 - command (1000000)
 - Bob (1000000)
Extras:
 - repetition: [[u'command', NatlinkDictationContainer(Alice, command, Bob)]]

"""


#---------------------------------------------------------------------------

import doctest
from dragonfly                          import *
from dragonfly.test.infrastructure      import RecognitionFailure
from dragonfly.test.element_testcase    import ElementTestCase
from dragonfly.test.element_tester      import ElementTester


#---------------------------------------------------------------------------

class RecognitionAnalysisRule(CompoundRule):
    """
        Base class that implements reporting in human-readable format
        details about the recognized phrase. It is used by the actual
        testing rules below, and allows the doctests above to be easily
        readable and informative.

    """

    def _process_recognition(self, node, extras):
        Paste(text).execute()

    def value(self, node):
        return self.get_recognition_info(node)

    def get_recognition_info(self, node):
        output = []
        output.append('Recognition: "{0}"'.format(" ".join(node.words())))
        output.append('Word and rule pairs: ("1000000" is "dgndictation")')
        for word, rule in node.full_results():
            output.append(" - {0} ({1})".format(word, rule))
        output.append("Extras:")
        for key in sorted(extra.name for extra in self.extras):
            extra_node = node.get_child_by_name(key)
            if extra_node:
                output.append(" - {0}: {1}".format(key, extra_node.value()))
        return "\n".join(output)


#---------------------------------------------------------------------------

class ConsecutiveDictationRule(RecognitionAnalysisRule):

    spec   = "consecutive <dictation1> <dictation2> <dictation3>"
    extras = [Dictation("dictation1"),
              Dictation("dictation2"),
              Dictation("dictation3")]


#---------------------------------------------------------------------------

class MixedDictationRule(RecognitionAnalysisRule):

    spec   = "mixed [<dictation1>] <dictation2> command <dictation3>"
    extras = [Dictation("dictation1"),
              Dictation("dictation2"),
              Dictation("dictation3")]


#---------------------------------------------------------------------------

class RepeatedDictationRule(RecognitionAnalysisRule):

    spec   = "<repetition>"
    extras = [Repetition(name="repetition",
                         child=Sequence([Literal("command"),
                                         Dictation()]))]


#---------------------------------------------------------------------------

def main():
    engine = get_engine()
    engine.connect()
    try:
        doctest.testmod(verbose=True)
    finally:
        engine.disconnect()

if __name__ == "__main__":
    main()



#
# This file  Is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Loader of language-dependent objects
============================================================================

This file implements the loader for language-dependent objects within
the dragonfly.language sub-package.  At runtime it attempts to figure
out what the current user's language is, and then loads the appropriate
objects.

Examples of language-dependent objects within Dragonfly are the Integer,
Digits, and Number classes.

"""

import sys
import logging
from ..grammar.rule_base       import Rule
from ..grammar.elements_basic  import RuleRef
from ..error                   import DragonflyError


#---------------------------------------------------------------------------
# Loader class for accessing language-specific object at runtime.

class LanguageSpecificLoader(object):

    _log = logging.getLogger("lang")


    #-----------------------------------------------------------------------
    # Mapping of language tags to module names.  Each of the module names
    #  should implement the appropriate classes for the given
    #  speaker language.

    language_map = {
                    "ar":  "dragonfly.language.ar",
                    "de":  "dragonfly.language.de",
                    "en":  "dragonfly.language.en",
                    "id":  "dragonfly.language.id",
                    "ms":  "dragonfly.language.ms",
                    "nl":  "dragonfly.language.nl",
                   }


    def __init__(self, language=None):
        self._language = language
        self._module = None

    def __getattr__(self, name):
        return self.get_attribute(name)

    def get_attribute(self, name):
        # Make sure the appropriate language-specific module has been
        #  loaded and is available.
        if not self._module:
            self._load_module()

        # Attempt to retrieve the requested attribute from the
        #  language-specific module.
        try:
            return getattr(self._module, name)
        except AttributeError:
            # Raise an error unless getting ShortIntegerContent.
            if name != "ShortIntegerContent":
                raise DragonflyError("Language %r does not implement %r."
                                     % (self._language, name))
            else:
                self._log.warning("Language %r does not implement %r, "
                                  "falling back on %r."
                                  %(self._language, name, "IntegerContent"))
                return getattr(self._module, "IntegerContent")

    def _load_module(self):
        if not self._language:
            self._language = self._get_engine_language()
        self._module = self._load_module_for_language(self._language)

    def _get_engine_language(self):
        try:
            from ..engines import get_engine
            language = get_engine().language
        except Exception as e:
            self._log.exception("Failed to retrieve speaker language: %s" % e)
            raise
        return language

    def _load_module_for_language(self, language):
        # Lookup the module name for the given language.
        try:
            module_name = self.language_map[language]
        except KeyError:
            raise DragonflyError("Attempt to load unknown language: %r"
                                 % language)

        # Attempt to import the language-dependent module.
        try:
            top_module = __import__(module_name)
        except ImportError as e:
            self._log.exception("Failed to load module %r for language %r"
                                % (module_name, language))
            raise
        try:
            module = sys.modules[module_name]
        except KeyError:
            self._log.exception("Failed to load module %r for language %r"
                                % (module_name, language))
            raise

        return module


#---------------------------------------------------------------------------
# Create single instance of language loader.
# This instance will be used by code which needs to access language-specific
#  objects, such as the Integer class which retrieves its language-specific
#  content using this language loader.

language = LanguageSpecificLoader()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

from .clipboard import BaseClipboard, Clipboard



"""
Example script for recognising speech from a wave file using the CMU Pocket
Sphinx dragonfly engine.

"""

import argparse

from dragonfly import get_engine


def main():
    desc = "Example script for using the " \
        "'SphinxEngine.process_wave_file' method"
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("file", type=argparse.FileType("r"),
                        help="A .wav file.")
    args = parser.parse_args()
    file_path = args.file.name
    args.file.close()

    # Set up the engine.
    # process_wave() handles START_ASLEEP = True (default), but a warning
    # message will be logged. We can suppress that here.
    engine = get_engine("sphinx")
    engine.config.START_ASLEEP = False

    # Voice activity detection config values may need to be adjusted to
    # recognise speech. Do this by changing the decoder's config.
    # Note that these values represent a number of audio frames, not time
    # intervals.
    # For example:
    engine.config.DECODER_CONFIG.set_int("-vad_postspeech", 30)

    # Process the wave file and print any results.
    # Note: process_wave_file() is a generator function.
    engine.connect()
    had_result = False
    for result in engine.process_wave_file(file_path):
        print(result)
        had_result = True

    # Exit with 1 if there were no results.
    if not had_result:
        exit(1)


if __name__ == "__main__":
    main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#


"""
Multiplexing interface for the CMU Pocket Sphinx engine
============================================================================

"""

from ..base import DelegateTimerManager


class SphinxTimerManager(DelegateTimerManager):
    """
    Timer manager for the CMU Pocket Sphinx engine.

    This class allows running timer functions if the engine is currently
    processing audio via one of three engine processing methods:

    - :meth:`process_buffer`
    - :meth:`process_wave_file`
    - :meth:`recognise_forever`

    Timer functions will run whether or not recognition is paused
    (i.e. in sleep mode).

    **Note**: long-running timers will block dragonfly from processing what
    was said, so be careful with how you use them! Audio frames will not
    normally be dropped because of timers, long-running or otherwise.

    Normal threads can be used instead of timers if desirable. This is
    because the main recognition loop is done in Python rather than in C/C++
    code, so there are no unusual multi-threading limitations.

    """



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#


import unittest
from six import PY2

from dragonfly.windows.window import Window


#===========================================================================

class TestWindow(unittest.TestCase):

    def setUp(self):
        pass

    def test_set_handle(self):
        """ Test access to Window.handle property. """

        # Verify that only integers and longs are accepted.
        Window(0)
        Window(int(1))
        if PY2:
            Window(long(2))
        self.assertRaises(TypeError, Window, [None])
        self.assertRaises(TypeError, Window, ["string"])
        self.assertRaises(TypeError, Window, [3.4])

#===========================================================================

if __name__ == "__main__":
    unittest.main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Date and time elements for the English language
============================================================================

"""

from datetime import date, time, timedelta
from ...grammar.elements  import Alternative, Compound, Choice
from ..base.integer       import Integer, IntegerRef


#---------------------------------------------------------------------------

#names = "January February March April May June July August September October November December".split()
#for index, name in enumerate(names):
#    print '%-13s %2d,' % ('"%s":' % name, index + 1)

month_names = {
               "January":     1,
               "February":    2,
               "March":       3,
               "April":       4,
               "May":         5,
               "June":        6,
               "July":        7,
               "August":      8,
               "September":   9,
               "October":    10,
               "November":   11,
               "December":   12,
              }

day_names   = {
               "Monday":      0,
               "Tuesday":     1,
               "Wednesday":   2,
               "Thursday":    3,
               "Friday":      4,
               "Saturday":    5,
               "Sunday":      6,
              }


#---------------------------------------------------------------------------

class Month(Choice):

    def __init__(self, name):
        Choice.__init__(self, name=name, choices=month_names)


class Day(Choice):

    def __init__(self, name):
        Choice.__init__(self, name=name, choices=day_names)


class Year(Alternative):

    alts = [
            IntegerRef("year", 2000, 2100),
            Compound(
                     spec="<century> <year>",
                     extras=[Integer("century", 20, 21),
                             IntegerRef("year", 10, 100)],
                     value_func=lambda n, e: e["century"] * 100 + e["year"]
                    ),
            Compound(
                     spec="<century> <year>",
                     extras=[Integer("century", 19, 20),
                             IntegerRef("year", 1, 100)],
                     value_func=lambda n, e: e["century"] * 100 + e["year"]
                    ),
           ]

    def __init__(self, name):
        Alternative.__init__(self, name=name, children=self.alts)


#---------------------------------------------------------------------------

class AbsoluteDate(Compound):

    spec = "(<day> <month> | <month> <day>) [<year>]"
    extras = [IntegerRef("day", 1, 32), Month("month"), Year("year")]

    def __init__(self, name):
        Compound.__init__(self, name=name, spec=self.spec,
                          extras=self.extras)

    def value(self, node):
        month      = node.get_child_by_name("month").value()
        day        = node.get_child_by_name("day").value()
        year_node  = node.get_child_by_name("year")
        if year_node is None:
            today = date.today()
            year = today.year
            if month - today.month > 6:
                # More than six months in the future, use last year.
                year -= 1
            elif month - today.month < -6:
                # More than six months in the past, use next year.
                year += 1
        else:
            year = year_node.value()
        return date(year, month, day)


class RelativeDate(Alternative):

    class _DayOffset(Choice):
        def __init__(self):
            choices = {
                       "<n> days ago": -1,
                       "yesterday":    -1,
                       "today":         0,
                       "tomorrow":     +1,
                       "in <n> days":  +1,
                      }
            extras = [IntegerRef("n", 1, 100)]
            Choice.__init__(self, name=None, choices=choices, extras=extras)

        def value(self, node):
            value = Choice.value(self, node)
            n = node.get_child_by_name("n")
            print("November:", n)
            if n is not None:
                value = value * n.value()
            return date.today() + timedelta(days=value)

    class _WeekdayOffset(Choice):
        def __init__(self):
            choices = {
                       "(last | past) <day>":  "last day",
                       "(next | this) <day>":  "next day",
                       "last week <day>":      "last week",
                       "next week <day>":      "next week",
                      }
            extras = [Day("day")]
            Choice.__init__(self, name=None, choices=choices, extras=extras)

        def value(self, node):
            value = Choice.value(self, node)
            day = node.get_child_by_name("day").value()
            now = date.today().weekday()
            print(value, day, now)
            if value == "last day":
                if day < now:  day_offset = -now + day
                else:          day_offset = -7 - now + day
            elif value == "next day":
                if day < now:  day_offset = 7 - now + day
                else:          day_offset = day - now
            elif value == "last week":
                day_offset = -now - 7 + day
            elif value == "next week":
                day_offset = -now + 7 + day
            return date.today() + timedelta(days=day_offset)

    alts = [
            _DayOffset(),
            _WeekdayOffset(),
           ]

    def __init__(self, name):
        Alternative.__init__(self, name=name, children=self.alts)


class Date(Alternative):

    alts = [
            AbsoluteDate(None),
            RelativeDate(None),
           ]

    def __init__(self, name):
        Alternative.__init__(self, name=name, children=self.alts)


#---------------------------------------------------------------------------

class MilitaryTime(Compound):

    spec = "<hour> (hundred | (oh | zero) <min_1_10> | <min_10_60>)"
    extras = [
              Integer("hour", 0, 25),
              IntegerRef("min_1_10", 1, 10),
              IntegerRef("min_10_60", 10, 60),
             ]

    def __init__(self, name):
        Compound.__init__(self, name=name, spec=self.spec,
                          extras=self.extras)

    def value(self, node):
        hour = node.get_child_by_name("hour").value()
        if node.has_child_with_name("min_1_10"):
            minute = node.get_child_by_name("min_1_10").value()
        elif node.has_child_with_name("min_10_60"):
            minute = node.get_child_by_name("min_10_60").value()
        else:
            minute = 0
        return time(hour, minute)


class Time(Alternative):

    alts = [
            MilitaryTime(None),
           ]

    def __init__(self, name):
        Alternative.__init__(self, name=name, children=self.alts)

#---------------------------------------------------------------------------




#
# This file is part of Dragonfly.
# (c) Copyright 2018 by Dane Finlay
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.

import argparse
import ast
import glob
import logging
import os
import re
import sys
import time

import six

from dragonfly import get_engine, MimicFailure, EngineError
from dragonfly.loader import CommandModule, CommandModuleDirectory

LOG = logging.getLogger("command")


#---------------------------------------------------------------------------
# CLI helper functions.

def _set_logging_level(args):
    if args.quiet:
        args.log_level = "WARNING"

    # Set up logging with the specified logging level.
    logging.basicConfig(level=getattr(logging, args.log_level))


def _init_engine(args):
    # Retrieve the engine option pairs from the arguments.
    # TODO Remove multiple engine options per -o arg in version 1.0.0.
    options = {}
    for argument in args.engine_options:
        for options_list in argument:
            for option, value in options_list:
                options[option] = value

    try:
        # Initialize the specified engine with options, catching and
        #  reporting errors.
        engine = get_engine(args.engine, **options)
    except EngineError as e:
        LOG.error(e)
        engine = None

    # Set the engine language if necessary.
    if engine and args.language != engine.language:
        try:
            engine.language = args.language
        except AttributeError:
            LOG.error("Cannot set language for engine %r", engine.name)
            engine = None

    return engine


def _load_cmd_modules(args):
    # Retrieve filenames from the arguments.
    files = []
    for lst in args.files:
        files.extend(lst)

    # Load each specified command module.  Errors during loading will be
    #  caught and logged.
    return_code = 0
    for f in files:
        module_ = CommandModule(f)
        module_.load()
        if not module_.loaded:
            return_code = 1

    # Return the overall success of module loading.
    return return_code


def _load_cmd_module_dirs(args):
    # Load command modules from each specified directory.  Errors during
    #  loading will be caught and logged.
    recursive = args.recursive
    return_code = 0
    for d in args.module_dirs:
        module_directory = CommandModuleDirectory(d, recursive=recursive)
        module_directory.load()
        if not module_directory.loaded:
            return_code = 1

    # Return the overall success of module loading.
    return return_code


def _on_begin():
    print("Speech start detected.")


def _on_recognition(words):
    message = u"Recognized: %s" % u" ".join(words)

    # This only seems to be an issue with Python 2.7 on Windows.
    if six.PY2:
        encoding = sys.stdout.encoding or "ascii"
        message = message.encode(encoding, errors='replace')
    print(message)


def _on_failure():
    print("Sorry, what was that?")


def _do_recognition(engine, args):
    # Start the engine's main recognition loop, registering recognition
    # callback functions only if necessary.
    try:
        if not args.no_recobs_messages:
            engine.do_recognition(_on_begin, _on_recognition, _on_failure)
        else:
            engine.do_recognition()
    except KeyboardInterrupt:
        pass


#---------------------------------------------------------------------------
# Main CLI functions.

def cli_cmd_test(args):
    # Set the logging level.
    _set_logging_level(args)

    # Initialise the specified engine. Return early if there was an error.
    engine = _init_engine(args)
    if engine is None:
        return 1

    # Connect to the engine, load command modules, take input from stdin and
    # disconnect from the engine if interrupted or if EOF is received.
    LOG.debug("Testing with engine '%s'", args.engine)
    with engine.connection():
        # Load each command module. Errors during loading will be caught and
        # logged. Use the overall success of module loading and/or mimic
        # calls as the return code.
        return_code = _load_cmd_modules(args)

        # Return early if --no-input was specified.
        if args.no_input:
            return return_code

        # Get the number of seconds to delay between each mimic() call
        # (default 0). Log a message if the delay is non-zero.
        delay = args.delay
        if delay:
            LOG.info("Calls to mimic() will be delayed by %.2f seconds "
                     "as specified", delay)

        # Read lines from stdin and pass them to engine.mimic. Strip excess
        # white space from each line. Report any mimic failures.
        LOG.info("Enter commands to mimic followed by new lines.")
        try:
            # Use iter to avoid a bug in Python 2.x:
            # https://bugs.python.org/issue3907
            for line in iter(sys.stdin.readline, ''):
                line = line.strip()
                if not line:  # skip empty lines.
                    continue

                # Delay between mimic() calls if necessary.
                if delay > 0:
                    time.sleep(delay)

                try:
                    engine.mimic(line.split())
                    LOG.info("Mimic success for words: %s", line)
                except MimicFailure:
                    LOG.error("Mimic failure for words: %s", line)
                    return_code = 1
        except KeyboardInterrupt:
            pass

    # Return the success of this command.
    return return_code


def cli_cmd_load(args):
    # Set the logging level.
    _set_logging_level(args)

    # Initialise the specified engine. Return early if there was an error.
    engine = _init_engine(args)
    if engine is None:
        return 1

    # Connect the engine, load each command module and start the main
    # recognition loop. The loop will normally exit on engine.disconnect()
    # or a keyboard interrupt.
    LOG.debug("Recognizing with engine '%s'", args.engine)
    with engine.connection():
        return_code = _load_cmd_modules(args)

        # Return early if --no-input was specified.
        if args.no_input:
            return return_code

        _do_recognition(engine, args)

    # Return the success of module loading.
    return return_code


def cli_cmd_load_directory(args):
    # Set the logging level.
    _set_logging_level(args)

    # Initialise the specified engine. Return early if there was an error.
    engine = _init_engine(args)
    if engine is None:
        return 1

    # Connect the engine, load each command module and start the main
    # recognition loop. The loop will normally exit on engine.disconnect()
    # or a keyboard interrupt.
    LOG.debug("Recognizing with engine '%s'", args.engine)
    with engine.connection():
        if args.recursive:
            LOG.info("Loading command modules in sub-directories as "
                     "specified (recursive mode).")
        return_code = _load_cmd_module_dirs(args)

        # Return early if --no-input was specified.
        if args.no_input:
            return return_code

        _do_recognition(engine, args)

    # Return the success of module loading.
    return return_code


_COMMAND_MAP = {
    "test": cli_cmd_test,
    "load": cli_cmd_load,
    "load-directory": cli_cmd_load_directory,
}


#---------------------------------------------------------------------------
# argparse helper functions.

def _build_argument(*args, **kwargs):
    return args, kwargs


def _add_arguments(parser, *arguments):
    for (args, kwargs) in arguments:
        parser.add_argument(*args, **kwargs)


def _smart_cast(value):
    """ Attempts to convert given str to a more precise type based on Python literals. """
    try:
        return ast.literal_eval(value)
    except:
        return value


def _engine_options_string(string):
    if '=' not in string:
        msg = "Invalid engine option: %r" % string
        raise argparse.ArgumentTypeError(msg)

    # Return a list of key/value arguments separated by commas or spaces.
    # TODO Remove multiple engine options per -o arg in version 1.0.0.
    options = []
    for sub_string in re.split('[,\\s]', string):
        if not sub_string:  # Filter out empty strings.
            continue

        # There must be one valid engine option per sub-string.
        parts = sub_string.split('=')
        if len(parts) != 2 or not (parts[0] and parts[1]):
            msg = "Invalid engine option: %r" % sub_string
            raise argparse.ArgumentTypeError(msg)

        arg = parts[0]
        value = _smart_cast(parts[1])
        options.append((arg, value))

    return options


def _filename(string):
    """ Same as argparse.FileType, but returns the filename. """
    f = argparse.FileType("r")(string)
    fname = f.name
    f.close()
    return fname


def _valid_filename_or_pattern(string):
    # Expand glob patterns or interpret the string as an existing file path.
    if "?" in string or "*" in string:
        files = glob.glob(string)
        if files:
            result = [_filename(f) for f in files]
        else:
            result = [_filename(string)]
    else:
        result = [_filename(string)]
    return result


def _valid_directory_path(string):
    if not os.path.isdir(string):
        msg = "No such directory: %r" % string
        raise argparse.ArgumentTypeError(msg)
    return string


#---------------------------------------------------------------------------
# Main argparse functions.

def make_arg_parser():
    # pylint: disable=too-many-locals
    # Suppress warnings about too many local variables in this function.

    parser = argparse.ArgumentParser(
        prog="python -m dragonfly",
        description="Command-line interface to the Dragonfly speech "
                    "recognition framework"
    )
    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    # Define common arguments.
    cmd_module_files_argument = _build_argument(
        "files", metavar="file", nargs="*", type=_valid_filename_or_pattern,
        help="Command module file(s)."
    )

    # TODO Rename to "--engine-option" in version 1.0.0.
    engine_options_argument = _build_argument(
        "-o", "--engine-options", default=[], nargs="+", action="append",
        type=_engine_options_string,
        help="One or more engine options to be passed to *get_engine()*. "
             "Each option should specify a key word argument and value. "
             "Multiple options should be separated by spaces or commas. "
             "Values are treated as Python literals if possible, "
             "otherwise as strings."
    )

    language_argument = _build_argument(
        "--language", default="en",
        help="Speaker language to use. Only applies if using an engine "
        "backend that supports changing the language (e.g. the \"text\" "
        "engine)."
    )
    no_input_argument = _build_argument(
        "-n", "--no-input", default=False, action="store_true",
        help="Whether to load command modules and then exit without "
             "reading input from stdin or recognizing speech."
    )
    log_level_argument = _build_argument(
        "-l", "--log-level", default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Log level to use."
    )
    quiet_argument = _build_argument(
        "-q", "--quiet", default=False, action="store_true",
        help="Equivalent to '-l WARNING' -- suppresses INFO and DEBUG "
             "logging."
    )

    # Create the parser for the "test" command.
    parser_test = subparsers.add_parser(
        "test",
        help="Load grammars from Python files for testing with a "
        "dragonfly engine. By default input from stdin is passed to "
        "engine.mimic() after command modules are loaded."
    )
    engine_argument = _build_argument(
        "-e", "--engine", default="text",
        help="Name of the engine to use for testing."
    )
    delay_argument = _build_argument(
        "-d", "--delay", default=0, type=float,
        help="Time in seconds to delay before mimicking each command. This "
        "is useful for testing contexts."
    )
    _add_arguments(
        parser_test,
        cmd_module_files_argument, engine_argument, engine_options_argument,
        language_argument, no_input_argument, delay_argument,
        log_level_argument, quiet_argument
    )

    # Define common arguments for the "load" and "load-directory" commands.
    engine_argument = _build_argument(
        "-e", "--engine", default=None,
        help="Name of the engine to use for loading and recognizing from "
             "command modules. By default, this is the first available "
             "engine backend."
    )
    no_recobs_messages_argument = _build_argument(
        "--no-recobs-messages", default=False, action="store_true",
        help="Disable recognition state messages for each spoken phrase."
    )

    # Create the parser for the "load" command.
    parser_load = subparsers.add_parser(
        "load", help="Load and recognize from command module files."
    )
    _add_arguments(
        parser_load,
        cmd_module_files_argument, engine_argument, engine_options_argument,
        language_argument, no_input_argument, no_recobs_messages_argument,
        log_level_argument, quiet_argument
    )

    # Create the parser for the "load-directory" command.
    parser_load_directory = subparsers.add_parser(
        "load-directory",
        help="Load and recognize from command module files in one or more"
             " directories.  Only module files starting with an underscore"
             " (_*.py) are loaded by this command."
    )
    module_dirs_argument = _build_argument(
        "module_dirs", metavar="dir", nargs="+", type=_valid_directory_path,
        help="One or more command module directories."
    )
    recursive_argument =  _build_argument(
        "-r", "--recursive", default=False, action="store_true",
        help="Whether to recursively load command modules in "
             "sub-directories."
    )
    _add_arguments(
        parser_load_directory,
        module_dirs_argument, recursive_argument, engine_argument,
        engine_options_argument, language_argument, no_input_argument,
        no_recobs_messages_argument, log_level_argument, quiet_argument
    )

    # Return the argument parser.
    return parser


def main():
    # Parse the arguments and get the relevant function. Exit if the command
    # is not implemented.
    args = make_arg_parser().parse_args()

    def not_implemented(_):
        print("Command '%s' is not implemented" % args.command)
        return 1

    func = _COMMAND_MAP.get(args.command, not_implemented)

    # Call the function and exit using the result.
    return_code = func(args)
    exit(return_code)


if __name__ == '__main__':
    main()
